import os

import numpy as np


def load_rxn_spectra():
    """
    Loads the raman spectral dataset of a reaction.

    Returns
    -------
    X : np.ndarray of shape (n_wavenumbers, n_samples)
        The spectral data.
    wv : np.ndarray of shape (n_wavenumbers,)
        The wavenumbers corresponding to the spectra.
    time : np.ndarray of shape (n_samples,)
        The time at which the samples were acquired.

    """
    DATA_DIR_PATH = os.path.join(os.path.dirname(__file__))

    fpath = os.path.join(DATA_DIR_PATH, "raman.csv")
    data = np.genfromtxt(fpath, delimiter=",", skip_header=1)

    time = np.genfromtxt(
        fpath, delimiter=",", max_rows=1, missing_values="", filling_values=np.nan
    )[1:]
    wv = data[:, 0].astype(dtype=np.float64)
    X = data[:, 1:].astype(dtype=np.float64)

    return X, wv, time



import numpy as np
from numpy.typing import NDArray
import pytest

from mcrnmf.nnls import FPGM, _initialize_H, _setup_input_args


def is_unimodal(arr_1d: NDArray) -> bool:
    """Verifies that the array has at most one peak"""
    increasing = False
    decreasing = False
    for i in range(1, arr_1d.size):
        if arr_1d[i] > arr_1d[i - 1]:
            increasing = True
            if decreasing:
                return False
        elif arr_1d[i] < arr_1d[i - 1]:
            decreasing = True
    return True


class TestFPGMValidation:
    """Test initialization parameter validation in FPGM."""

    def test_constraint_kind_validation(self):
        """Test constraint_kind parameter validation."""
        # Type validation
        msg = r"`constraint_kind` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            FPGM(constraint_kind="1")
        with pytest.raises(TypeError, match=msg):
            FPGM(constraint_kind=1.0)

        # Value validation
        msg = r"`constraint_kind` must be in 0, 1, 2, or 3."
        with pytest.raises(ValueError, match=msg):
            FPGM(constraint_kind=4)
        with pytest.raises(ValueError, match=msg):
            FPGM(constraint_kind=-1)

    def test_iter_max_validation(self):
        """Test iter_max parameter validation."""
        # Type validation
        msg = r"`iter_max` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            FPGM(iter_max="200")
        with pytest.raises(TypeError, match=msg):
            FPGM(iter_max=200.0)

        # Value validation
        msg = r"`iter_max` must be > 0"
        with pytest.raises(ValueError, match=msg):
            FPGM(iter_max=0)
        with pytest.raises(ValueError, match=msg):
            FPGM(iter_max=-10)

    def test_tol_validation(self):
        """Test tol parameter validation."""
        # Type validation
        msg = r"`tol` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            FPGM(tol=1)
        with pytest.raises(TypeError, match=msg):
            FPGM(tol="0.001")

        # Value validation
        msg = r"`tol` must be in the open interval \(0, 1\)"
        with pytest.raises(ValueError, match=msg):
            FPGM(tol=0.0)
        with pytest.raises(ValueError, match=msg):
            FPGM(tol=1.0)
        with pytest.raises(ValueError, match=msg):
            FPGM(tol=2.0)


class TestSetupInputArgs:
    """Test the validation in _setup_input_args function."""

    @pytest.fixture
    def valid_matrices(self):
        """Create valid matrices for testing."""
        rng = np.random.default_rng(42)
        X = rng.random((20, 10)).astype(np.float64)
        W = rng.random((20, 5)).astype(np.float64)
        H = rng.random((5, 10)).astype(np.float64)
        return X, W, H

    def test_input_type_validation(self, valid_matrices):
        """Test validation of input types."""
        X, W, H = valid_matrices

        # X must be ndarray
        with pytest.raises(TypeError, match=r"`X` must be a numpy array."):
            _setup_input_args(X.tolist(), W, H)

        # W must be ndarray
        with pytest.raises(TypeError, match=r"`W` must be a numpy array."):
            _setup_input_args(X, W.tolist(), H)

        # H must be ndarray or None
        with pytest.raises(TypeError, match=r"`H` must be a numpy array or None."):
            _setup_input_args(X, W, "not_array")

    def test_shape_validation(self, valid_matrices):
        """Test validation of matrix shapes."""
        X, W, H = valid_matrices

        # W rows must match X rows
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: rows of `W`: \d+ not equal to rows of X: \d+",
        ):
            _setup_input_args(X, W[1:, :], H)

        # H rows must match W columns
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: rows of `H`: \d+ not equal to columns of W: \d+",
        ):
            _setup_input_args(X, W, H[1:, :])

        # H cols must match X cols
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: columns of `H`: \d+ not equal to columns of X: \d+",
        ):
            _setup_input_args(X, W, H[:, 1:])

    def test_nonnegativity_validation(self, valid_matrices):
        """Test validation of non-negativity constraints."""
        X, W, H = valid_matrices

        # H must be non-negative
        H_neg = H.copy()
        H_neg[0, 0] = -0.1
        with pytest.raises(ValueError, match=r"All elements of `H` must be >= 0."):
            _setup_input_args(X, W, H_neg)

    def test_H_initialization(self, valid_matrices):
        """Test H initialization when None is provided."""
        X, W, _ = valid_matrices

        # When H is None, it should be initialized
        X_out, W_out, H_out = _setup_input_args(X, W, None)

        # Check that shapes are correct
        assert H_out.shape == (W.shape[1], X.shape[1])

        # Verify contiguous arrays with float64 dtype
        assert X_out.flags.c_contiguous
        assert W_out.flags.c_contiguous
        assert H_out.flags.c_contiguous
        assert X_out.dtype == np.float64
        assert W_out.dtype == np.float64
        assert H_out.dtype == np.float64

        # Check that H is non-negative
        assert np.all(H_out >= 0)

    def test_valid_inputs_pass_validation(self, valid_matrices):
        """Test that valid inputs pass validation without errors."""
        X, W, H = valid_matrices
        X_out, W_out, H_out = _setup_input_args(X, W, H)

        # Check that shapes are preserved
        assert X_out.shape == X.shape
        assert W_out.shape == W.shape
        assert H_out.shape == H.shape


class TestInitializeH:
    """Test the H initialization function."""

    @pytest.fixture
    def matrices(self):
        """Create matrices for testing initialization."""
        rng = np.random.default_rng(42)
        X = rng.random((20, 10)).astype(np.float64)
        W = rng.random((20, 5)).astype(np.float64)
        return X, W

    def test_initialize_H_shape(self, matrices):
        """Test that initialized H has the correct shape."""
        X, W = matrices
        H = _initialize_H(X, W)
        assert H.shape == (W.shape[1], X.shape[1])

    def test_initialize_H_nonnegativity(self, matrices):
        """Test that initialized H is non-negative."""
        X, W = matrices
        H = _initialize_H(X, W)
        assert np.all(H >= 0)

    def test_initialize_H_warning(self):
        """Test warning for ill-conditioned W."""
        # Create a matrix W with high condition number
        X = np.ones((5, 3))
        W = np.ones((5, 2))  # Rank deficient

        with pytest.warns(UserWarning, match=r"`W` has very high condition number.*"):
            H = _initialize_H(X, W)


class TestApplyProjection:
    """Test the _apply_projection method of FPGM."""

    @pytest.fixture
    def test_matrix(self):
        """Create a test matrix for projection."""
        rng = np.random.default_rng(42)
        A = rng.random((3, 10)).astype(np.float64)
        A[0, 0] = -0.5  # Add a negative value
        return A

    def test_projection_constraint_0(self, test_matrix):
        """Test projection with constraint_kind=0."""
        fpgm = FPGM(constraint_kind=0)
        A = test_matrix.copy()
        fpgm._apply_projection(A, 0)

        # Check that all values are non-negative
        assert np.all(A >= 0)
        # Check that only negative values were modified
        assert np.isclose(A[0, 0], 0)

    def test_projection_constraint_1(self, test_matrix):
        """Test projection with constraint_kind=1."""
        fpgm = FPGM(constraint_kind=1)
        A = test_matrix.copy()
        fpgm._apply_projection(A, 1)

        # Check that all values are non-negative
        assert np.all(A >= 0)
        # Check that column sums are <= 1
        assert np.all(np.sum(A, axis=0) <= 1.0 + 1e-10)

    def test_projection_constraint_2(self, test_matrix):
        """Test projection with constraint_kind=2."""
        fpgm = FPGM(constraint_kind=2)
        A = test_matrix.copy()
        fpgm._apply_projection(A, 2)

        # Check that all values are non-negative
        assert np.all(A >= 0)
        # Check that row sums equal 1
        assert np.allclose(np.sum(A, axis=1), 1.0, atol=1e-10)

    def test_projection_constraint_3(self, test_matrix):
        """Test projection with constraint_kind=3."""
        fpgm = FPGM(constraint_kind=3)
        A = test_matrix.copy()
        fpgm._apply_projection(A, 3)

        # Check that all values are non-negative
        assert np.all(A >= 0)
        # Check that column sums equal 1
        assert np.allclose(np.sum(A, axis=0), 1.0, atol=1e-10)


class TestSolve:
    """Test the solve method of FPGM."""

    @pytest.fixture
    def test_data(self):
        """Create test data for solve."""
        rng = np.random.default_rng(42)
        X = rng.random((20, 10)).astype(np.float64)
        W = rng.random((20, 5)).astype(np.float64)
        H = rng.random((5, 10)).astype(np.float64)
        return X, W, H

    @pytest.mark.parametrize("constraint_kind", [0, 1, 2, 3])
    def test_solve_constraints(self, test_data, constraint_kind):
        """Test that solve respects different constraints."""
        X, W, _ = test_data
        fpgm = FPGM(constraint_kind=constraint_kind, iter_max=10)
        H, _ = fpgm.solve(X, W)

        # Check that solution is non-negative
        assert np.all(H >= 0)

        if constraint_kind == 1:
            # Check that column sums are <= 1
            assert np.all(np.sum(H, axis=0) <= 1.0 + 1e-10)
        elif constraint_kind == 2:
            # Check that row sums equal 1
            assert np.allclose(np.sum(H, axis=1), 1.0, atol=1e-10)
        elif constraint_kind == 3:
            # Check that column sums equal 1
            assert np.allclose(np.sum(H, axis=0), 1.0, atol=1e-10)

    def test_solve_with_H(self, test_data):
        """Test solve with provided initial H."""
        X, W, H = test_data
        fpgm = FPGM(iter_max=10)
        H_out, _ = fpgm.solve(X, W, H)

        # Check that output has same shape as input
        assert H_out.shape == H.shape
        # Check that H was modified (not identical to input)
        assert not np.array_equal(H_out, H)

    def test_solve_convergence(self, test_data):
        """Test solve convergence flag."""
        X, W, _ = test_data

        # With strict tolerance and few iterations, should not converge
        fpgm_no_converge = FPGM(iter_max=1, tol=1e-10)
        _, converged_1 = fpgm_no_converge.solve(X, W)
        assert converged_1 is False

        # With relaxed tolerance and more iterations, should converge
        fpgm_converge = FPGM(iter_max=100, tol=1e-2)
        _, converged_2 = fpgm_converge.solve(X, W)
        assert converged_2 is True

    def test_solve_objective_decreases(self, test_data):
        """Test that the objective function decreases during solve."""
        X, W, H_init = test_data

        # Calculate initial error
        initial_err = np.linalg.norm(X - W @ H_init, "fro")

        # Solve with initial H
        fpgm = FPGM(iter_max=50)
        H_final, _ = fpgm.solve(X, W, H_init.copy())

        # Calculate final error
        final_err = np.linalg.norm(X - W @ H_final, "fro")

        # Check that error decreased
        assert final_err <= initial_err


class TestSolveKnownH:
    @pytest.fixture
    def test_data(self):
        rng = np.random.default_rng(0)
        X = rng.random((8, 6)).astype(np.float64)
        W = rng.random((8, 4)).astype(np.float64)
        return X, W

    def test_known_H_type_and_shape(self, test_data):
        X, W = test_data
        fpgm = FPGM()
        with pytest.raises(
            TypeError, match=r"`known_H` must be None or a numpy array."
        ):
            fpgm.solve(X, W, known_H=123)
        bad_shape = np.empty((W.shape[1] + 1, X.shape[1]))
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: number of columns of `known_H` not equal to number "
            "of columns of `W`",
        ):
            fpgm.solve(X, W, known_H=bad_shape)

    def test_known_H_all_specified(self, test_data):
        X, W = test_data
        full = np.zeros((W.shape[1], X.shape[1]))
        with pytest.raises(
            ValueError, match=r"All values of `H` are specified as known, can't solve"
        ):
            fpgm = FPGM()
            fpgm.solve(X, W, known_H=full)

    def test_known_H_negative_finite(self, test_data):
        X, W = test_data
        known = np.full((W.shape[1], X.shape[1]), np.nan)
        known[0, 0] = -0.1
        fpgm = FPGM()
        with pytest.raises(
            ValueError, match=r"All finite elements of `known_H` must be >= 0"
        ):
            fpgm.solve(X, W, known_H=known)

    def test_known_H_no_finite_warns_and_ignores(self, test_data):
        X, W = test_data
        known = np.full((W.shape[1], X.shape[1]), np.nan)
        fpgm = FPGM(iter_max=5)
        with pytest.warns(
            UserWarning,
            match=r"`known_H` contains no finite values, setting it to None.",
        ):
            H_out, _ = fpgm.solve(X, W, known_H=known)
        # should behave like no known_H
        assert H_out.shape == (W.shape[1], X.shape[1])
        assert np.all(H_out >= 0)

    def test_solve_with_known_H_values_fixed(self, test_data):
        X, W = test_data
        H_init = np.zeros((W.shape[1], X.shape[1]))
        known = np.full_like(H_init, np.nan)
        known[1, 2] = 0.5
        known[3, 4] = 1.0
        fpgm = FPGM(iter_max=20)
        H_out, _ = fpgm.solve(X, W, known_H=known)
        assert np.isclose(H_out[1, 2], 0.5)
        assert np.isclose(H_out[3, 4], 1.0)
        # other entries free and non-negative
        mask = np.isnan(known)
        assert np.all(H_out[mask] >= 0)


class TestSolveUnimodalH:
    @pytest.fixture
    def test_data(self):
        rng = np.random.default_rng(1)
        X = rng.random((7, 9)).astype(np.float64)
        W = rng.random((7, 3)).astype(np.float64)
        return X, W

    def test_unimodal_H_type_and_length(self, test_data):
        X, W = test_data
        fpgm = FPGM()
        with pytest.raises(
            TypeError, match=r"`unimodal_H` must be of type bool or list of bool"
        ):
            fpgm.solve(X, W, unimodal_H=123)
        with pytest.raises(
            TypeError,
            match=r"If `unimodal_H` is a list, it must contain only boolean elements",
        ):
            fpgm.solve(X, W, unimodal_H=[True, 1, False])
        with pytest.raises(
            ValueError,
            match="If `unimodal_H` is specified as list, then the length of list must "
            "be equal to number of columns of W",
        ):
            fpgm.solve(X, W, unimodal_H=[True] * (W.shape[1] + 1))

    def test_solve_unimodal_all_rows(self, test_data):
        X, W = test_data
        fpgm = FPGM(iter_max=30)
        H_out, _ = fpgm.solve(X, W, unimodal_H=True)
        for row in range(W.shape[1]):
            assert is_unimodal(H_out[row, :])

    def test_solve_unimodal_selected_rows(self, test_data):
        X, W = test_data
        mask = [True, False, True]
        fpgm = FPGM(iter_max=30)
        H_out, _ = fpgm.solve(X, W, unimodal_H=mask)
        assert is_unimodal(H_out[0, :])
        assert is_unimodal(H_out[2, :])
        assert H_out.shape == (W.shape[1], X.shape[1])



import numpy as np
import pytest

from mcrnmf.models import FroALS, FroFPGM, MinVol, SNPA


class TestBaseNMFValidation:
    """Test initialization parameter validation in _BaseNMF."""

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_rank_validation(self, model_class):
        """Test rank parameter validation."""
        # type validation
        msg = r".*`rank` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            model_class(rank="2")
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2.0)

        # value validation
        msg = "`rank` must be > 0"
        with pytest.raises(ValueError, match=msg):
            model_class(rank=0)
        with pytest.raises(ValueError, match=msg):
            model_class(rank=-3)

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_constraint_kind_validation(self, model_class):
        """Test constraint_kind parameter validation"""
        # type validation
        msg = r".*`constraint_kind` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, constraint_kind="1")
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, constraint_kind=1.0)

        # value validation
        msg = r"`constraint_kind` must be 0, 1, 2, 3, or 4"
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, constraint_kind=5)

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_unimodal_constraints_validation(self, model_class):
        """Test unimodal_constraints parameter validation."""
        # type validation
        msg = r"`unimodal` must be None or a dictionary"
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, unimodal="dict")
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, unimodal=1)

        # empty dictionary validation
        msg = r"If not None, unimodal dictionary cannot be empty; must contain 'W' "
        "and/or 'H' keys."
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, unimodal={})

        # invalid key validation
        msg = r"`unimodal` dictionary can contain only keys in: \['W', 'H'\]"
        with pytest.raises(KeyError, match=msg):
            model_class(rank=2, unimodal={"W": True, "invalid_key": False})

        # value type validation
        msg = r"Unimodal constraint must be bool or list of booleans"
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, unimodal={"W": "True"})
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, unimodal={"H": 1})

        # list validation
        msg = r"Unimodal constraint list must contain only booleans"
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, unimodal={"W": [True, 1]})

        # list length validation
        msg = r"Unimodal constraint list must have length 2"
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, unimodal={"H": [True, False, True]})

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_iter_max_validation(self, model_class):
        """Test iter_max parameter validation."""
        # type validation
        msg = r".*`iter_max` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, iter_max="200")
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, iter_max=200.0)

        # value validation
        msg = r"`iter_max` must be >= 10"
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, iter_max=0)
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, iter_max=-200)

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_tol_validation(self, model_class):
        """Test tol parameter validation."""
        # Type validation
        msg = r".*`tol` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, tol=1)
        with pytest.raises(TypeError, match=msg):
            model_class(rank=2, tol="0.001")

        # Value validation
        msg = r"`tol` must be in the open interval \(0, 1\)"
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, tol=0.0)
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, tol=1.0)
        with pytest.raises(ValueError, match=msg):
            model_class(rank=2, tol=2.0)


class TestFitArgumentsValidation:
    """Test validation of fit arguments in _BaseNMF._validate_fit_args."""

    @pytest.fixture(params=[FroALS, FroFPGM, MinVol])
    def test_model(self, request):
        """Create model instances for testing validation."""
        model_class = request.param
        return model_class(rank=3)

    @pytest.fixture
    def valid_matrices(self):
        """Create valid matrices for fit."""
        X = np.random.rand(20, 10).astype(np.float64)
        Wi = np.random.rand(20, 3).astype(np.float64)
        Hi = np.random.rand(3, 10).astype(np.float64)
        return X, Wi, Hi

    def test_input_type_validation(self, test_model, valid_matrices):
        """Test validation of input types."""
        X, Wi, Hi = valid_matrices

        # X must be ndarray
        with pytest.raises(TypeError, match=r"`X` must be a numpy array"):
            test_model._validate_fit_args(X.tolist(), Wi, Hi, None, None)

        # Wi must be ndarray
        with pytest.raises(TypeError, match=r"`Wi` must be a numpy array"):
            test_model._validate_fit_args(X, Wi.tolist(), Hi, None, None)

        # Hi must be ndarray
        with pytest.raises(TypeError, match=r"`Hi` must be a numpy array"):
            test_model._validate_fit_args(X, Wi, Hi.tolist(), None, None)

        # known_W must be ndarray or None
        with pytest.raises(TypeError, match=r"`known_W` must be None or a numpy array"):
            test_model._validate_fit_args(X, Wi, Hi, "not_array", None)

        # known_H must be ndarray or None
        with pytest.raises(TypeError, match=r"`known_H` must be None or a numpy array"):
            test_model._validate_fit_args(X, Wi, Hi, None, "not_array")

    def test_shape_validation(self, test_model, valid_matrices):
        """Test validation of matrix shapes."""
        X, Wi, Hi = valid_matrices

        # Wi rows must match X rows
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: number of rows of `Wi` not equal to number of rows "
            "of `X`",
        ):
            test_model._validate_fit_args(X, Wi[1:, :], Hi, None, None)

        # Wi cols must match rank
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: number of columns of `Wi` not equal to `rank`",
        ):
            test_model._validate_fit_args(X, Wi[:, :2], Hi, None, None)

        # Hi rows must match rank
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: number of rows of `Hi` not equal to `rank`",
        ):
            test_model._validate_fit_args(X, Wi, Hi[:2, :], None, None)

        # Hi cols must match X cols
        with pytest.raises(
            ValueError,
            match=r"Shape mismatch: number of columns of `Hi` not equal to number of "
            "columns of `X`",
        ):
            test_model._validate_fit_args(X, Wi, Hi[:, :9], None, None)

        # known_W shape must match Wi shape
        known_W_wrong_shape = np.full((Wi.shape[0], Wi.shape[1] - 1), np.nan)
        with pytest.raises(
            ValueError, match=r"Shape mismatch: `known_W` shape not equal to `Wi` shape"
        ):
            test_model._validate_fit_args(X, Wi, Hi, known_W_wrong_shape, None)

        # known_H shape must match Hi shape
        known_H_wrong_shape = np.full((Hi.shape[0] - 1, Hi.shape[1]), np.nan)
        with pytest.raises(
            ValueError, match=r"Shape mismatch: `known_H` shape not equal to `Hi` shape"
        ):
            test_model._validate_fit_args(X, Wi, Hi, None, known_H_wrong_shape)

    def test_dtype_validation(self, test_model, valid_matrices):
        """Test validation of data types."""
        X, Wi, Hi = valid_matrices

        # X must be float64
        with pytest.raises(TypeError, match=r"The dtype of `X` elements should be"):
            test_model._validate_fit_args(X.astype(np.float32), Wi, Hi, None, None)

        # Wi must be float64
        with pytest.raises(TypeError, match=r"The dtype of `Wi` elements should be"):
            test_model._validate_fit_args(X, Wi.astype(np.float32), Hi, None, None)

        # Hi must be float64
        with pytest.raises(TypeError, match=r"The dtype of `Hi` elements should be"):
            test_model._validate_fit_args(X, Wi, Hi.astype(np.float32), None, None)

    def test_nonnegativity_validation(self, test_model, valid_matrices):
        """Test validation of non-negativity constraints."""
        X, Wi, Hi = valid_matrices

        # X must be non-negative
        X_neg = X.copy()
        X_neg[0, 0] = -0.1
        with pytest.raises(ValueError, match=r"All elements of `X` must be >= 0"):
            test_model._validate_fit_args(X_neg, Wi, Hi, None, None)

        # Wi must be non-negative
        Wi_neg = Wi.copy()
        Wi_neg[0, 0] = -0.1
        with pytest.raises(ValueError, match=r"All elements of `Wi` must be >= 0"):
            test_model._validate_fit_args(X, Wi_neg, Hi, None, None)

        # Hi must be non-negative
        Hi_neg = Hi.copy()
        Hi_neg[0, 0] = -0.1
        with pytest.raises(ValueError, match=r"All elements of `Hi` must be >= 0"):
            test_model._validate_fit_args(X, Wi, Hi_neg, None, None)

    def test_known_W_validation(self, test_model, valid_matrices):
        """Test validation of known_W."""
        X, Wi, Hi = valid_matrices

        # All NaNs in known_W should trigger warning
        known_W_all_nan = np.full(Wi.shape, np.nan)
        with pytest.warns(
            UserWarning,
            match=r"`known_W` contains no finite values, setting it to None",
        ):
            test_model._validate_fit_args(X, Wi, Hi, known_W_all_nan, None)

        # All finite values in known_W should raise error
        known_W_all_finite = np.ones_like(Wi)
        with pytest.raises(ValueError, match=r"All values of `W` are known.*"):
            test_model._validate_fit_args(X, Wi, Hi, known_W_all_finite, None)

        # Negative finite values in known_W should raise error
        known_W_with_neg = np.full(Wi.shape, np.nan)
        known_W_with_neg[0, 0] = -0.1
        with pytest.raises(
            ValueError, match=r"All finite elements of `known_W` must be >= 0"
        ):
            test_model._validate_fit_args(X, Wi, Hi, known_W_with_neg, None)

    def test_known_H_validation(self, test_model, valid_matrices):
        """Test validation of known_H."""
        X, Wi, Hi = valid_matrices

        # All NaNs in known_H should trigger warning
        known_H_all_nan = np.full(Hi.shape, np.nan)
        with pytest.warns(
            UserWarning,
            match=r"`known_H` contains no finite values, setting it to None",
        ):
            test_model._validate_fit_args(X, Wi, Hi, None, known_H_all_nan)

        # All finite values in known_H should raise error
        known_H_all_finite = np.ones_like(Hi)
        with pytest.raises(ValueError, match=r"All values of `H` are known.*"):
            test_model._validate_fit_args(X, Wi, Hi, None, known_H_all_finite)

        # Negative finite values in known_H should raise error
        known_H_with_neg = np.full(Hi.shape, np.nan)
        known_H_with_neg[0, 0] = -0.1
        with pytest.raises(
            ValueError, match=r"All finite elements of `known_H` must be >= 0"
        ):
            test_model._validate_fit_args(X, Wi, Hi, None, known_H_with_neg)

    def test_valid_inputs_pass_validation(self, test_model, valid_matrices):
        """Test that valid inputs pass validation without errors."""
        X, Wi, Hi = valid_matrices

        # Valid inputs with no known_W or known_H
        test_model._validate_fit_args(X, Wi, Hi, None, None)

        # Valid inputs with partially specified known_W and known_H
        known_W = np.full(Wi.shape, np.nan)
        known_W[0, 0] = 0.5

        known_H = np.full(Hi.shape, np.nan)
        known_H[0, 0] = 0.5

        test_model._validate_fit_args(X, Wi, Hi, known_W, known_H)


class TestBaseNMFFunctionality:
    """Test common fit functionality across NMF implementations."""

    @pytest.fixture
    def test_data(self):
        """Create test data that can be used across tests."""
        rng = np.random.default_rng(seed=42)
        X = rng.random((50, 30))
        rank = 3
        Wi = rng.random((50, rank))
        Hi = rng.random((rank, 30))
        return {"X": X, "Wi": Wi, "Hi": Hi, "rank": rank}

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_property_validation(self, test_data, model_class):
        """Test that validation of properties common to all NMF implementations."""
        model: FroALS | FroFPGM | MinVol = model_class(rank=test_data["rank"])

        with pytest.raises(AttributeError, match=r"`W` is not available.*"):
            _ = model.W

        with pytest.raises(AttributeError, match=r"`H` is not available.*"):
            _ = model.H

        with pytest.raises(AttributeError, match=r"`rel_loss_ls` is not available.*"):
            _ = model.rel_loss_ls

        with pytest.raises(AttributeError, match=r"`is_converged` is not available.*"):
            _ = model.is_converged

        # not truly a base nmf functionality but including it here for convenience
        with pytest.raises(
            AttributeError, match=r"`rel_reconstruction_error_ls` is not available.*"
        ):
            _ = model.rel_reconstruction_error_ls

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_basic_fit(self, test_data, model_class):
        """Test the basic fit method works and produces the expected attributes."""
        model: FroALS | FroFPGM | MinVol = model_class(rank=test_data["rank"])
        model.fit(X=test_data["X"], Wi=test_data["Wi"], Hi=test_data["Hi"])

        # check dimensions of resulting matrices
        assert model._W.shape == test_data["Wi"].shape
        assert model._H.shape == test_data["Hi"].shape

        # check that error decreases or stays the same
        assert model._rel_loss_ls[-1] <= model._rel_loss_ls[0]

    @pytest.mark.parametrize(
        "model_class, constraint_kind",
        [
            (FroALS, 0),
            (FroALS, 1),
            (FroALS, 2),
            (FroALS, 3),
            (FroALS, 4),
            (FroFPGM, 0),
            (FroFPGM, 1),
            (FroFPGM, 2),
            (FroFPGM, 3),
            (FroFPGM, 4),
            (MinVol, 0),
            (MinVol, 1),
            (MinVol, 2),
            (MinVol, 3),
            (MinVol, 4),
        ],
    )
    def test_constraints(self, test_data, model_class, constraint_kind):
        """Test that constraints are enforced correctly."""
        model: FroALS | FroFPGM | MinVol = model_class(
            rank=test_data["rank"], constraint_kind=constraint_kind
        )
        model.fit(X=test_data["X"], Wi=test_data["Wi"], Hi=test_data["Hi"])

        # check non-negativity constraint is always enforced
        assert np.all(model.W >= 0)
        assert np.all(model.H >= 0)

        if constraint_kind == 1:
            # simplex constraint on H
            assert np.all(np.sum(model.H, axis=0) <= 1.0 + 1e-10)
        elif constraint_kind == 2:
            # row simplex constraint on H
            assert np.allclose(np.sum(model.H, axis=1), 1.0, atol=1e-10)
        elif constraint_kind == 3:
            # column simplex constraint on W
            assert np.allclose(np.sum(model.W, axis=0), 1.0, atol=1e-10)
        elif constraint_kind == 4:
            # column simplex constraint on H
            assert np.allclose(np.sum(model.H, axis=0), 1.0, atol=1e-10)

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_unimodal_constraints(self, test_data, model_class):
        """Test that unimodal constraints are enforced correctly."""
        model: FroALS | FroFPGM | MinVol = model_class(
            rank=test_data["rank"], unimodal={"H": [True, False, True]}
        )
        model.fit(X=test_data["X"], Wi=test_data["Wi"], Hi=test_data["Hi"])

        def is_unimodal(arr_1d: np.ndarray):
            """Verifies that the array has at most one peak"""
            increasing = False
            decreasing = False
            for i in range(1, arr_1d.size):
                if arr_1d[i] > arr_1d[i - 1]:
                    increasing = True
                    if decreasing:
                        return False
                elif arr_1d[i] < arr_1d[i - 1]:
                    decreasing = True
            return True

        # assert first and third row are unimodal
        assert is_unimodal(model.H[0, :])
        assert is_unimodal(model.H[2, :])

    @pytest.mark.parametrize("model_class", [FroALS, FroFPGM, MinVol])
    def test_known_values_preserved(self, test_data, model_class):
        """Test that known values in W and H are preserved during updates."""

        # Create known_W and known_H with some specified values
        known_W = np.full_like(test_data["Wi"].copy(), np.nan)
        known_W[0, 0] = 0.5

        known_H = np.full_like(test_data["Hi"].copy(), np.nan)
        known_H[0, 0] = 0.7

        model = model_class(rank=test_data["rank"])
        model.fit(
            X=test_data["X"],
            Wi=test_data["Wi"],
            Hi=test_data["Hi"],
            known_W=known_W,
            known_H=known_H,
        )

        # Check that the known values are preserved
        assert model.W[0, 0] == 0.5
        assert model.H[0, 0] == 0.7


class TestSNPA:
    """Test all aspects of SNPA class."""

    @pytest.fixture
    def test_data(self):
        rng = np.random.default_rng(seed=42)
        X = rng.random((50, 30))
        rank = 3
        return {"X": X, "rank": rank}

    def test_rank_validation(self):
        """Test rank parameter validation."""
        # type validation
        msg = r".*`rank` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            SNPA(rank="2")
        with pytest.raises(TypeError, match=msg):
            SNPA(rank=2.0)

        # value validation
        msg = "`rank` must be > 0"
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=0)
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=-3)

    def test_iter_max_validation(self):
        """Test iter_max parameter validation."""
        # type validation
        msg = r".*`iter_max` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            SNPA(rank=2, iter_max="200")
        with pytest.raises(TypeError, match=msg):
            SNPA(rank=2, iter_max=200.0)

        # value validation
        msg = r"`iter_max` must be > 0"
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=2, iter_max=0)
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=2, iter_max=-200)

    def test_tol_validation(self):
        """Test tol parameter validation."""
        # Type validation
        msg = r".*`tol` must be of type.*"
        with pytest.raises(TypeError, match=msg):
            SNPA(rank=2, tol=1)
        with pytest.raises(TypeError, match=msg):
            SNPA(rank=2, tol="0.001")

        # Value validation
        msg = r"`tol` must be in the open interval \(0, 1\)"
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=2, tol=0.0)
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=2, tol=1.0)
        with pytest.raises(ValueError, match=msg):
            SNPA(rank=2, tol=2.0)

    def test_fit_X_validation(self, test_data):
        """Test X array validation."""
        # X must be ndarray
        with pytest.raises(TypeError, match=r"`X` must be a numpy array"):
            snpa = SNPA(rank=test_data["rank"])
            snpa.fit(X=test_data["X"].tolist())

        # Each element of X must of np.float64
        with pytest.raises(TypeError, match=r"The dtype of `X` elements should be.*"):
            snpa = SNPA(rank=test_data["rank"])
            snpa.fit(X=test_data["X"].astype(np.float32))

        # test nonnegativity
        with pytest.raises(ValueError, match=r"All elements of `X` must be >= 0"):
            snpa = SNPA(rank=test_data["rank"])
            temp_X = test_data["X"].copy()
            temp_X[0, 0] = -0.1
            snpa.fit(X=temp_X)

        # The number of columns in X cannot be lesser than rank
        with pytest.raises(
            ValueError, match=r"`rank` cannot be greater than columns in `X`*"
        ):
            snpa = SNPA(rank=test_data["rank"])
            snpa.fit(X=test_data["X"][:, :2])

    def test_tol_warning(self, test_data):
        """Test if tolerance warning is raised."""
        with pytest.warns(UserWarning, match=r"Extracted only.*"):
            snpa = SNPA(rank=test_data["rank"], tol=0.9)
            snpa.fit(X=test_data["X"])

    def test_property_validation(self, test_data):
        """Test that validation of properties."""
        snpa = SNPA(rank=test_data["rank"])

        with pytest.raises(AttributeError, match=r"'W' is not available.*"):
            _ = snpa.W

        with pytest.raises(AttributeError, match=r"'H' is not available.*"):
            _ = snpa.H

        with pytest.raises(AttributeError, match=r"'col_indices_ls' is not available."):
            _ = snpa.col_indices_ls

    def test_fit_functionality(self, test_data):
        """Test common SNPA fit functionality."""

        snpa = SNPA(rank=test_data["rank"])
        snpa.fit(X=test_data["X"])

        # check dimensions of resulting matrices
        assert snpa.W.shape == (test_data["X"].shape[0], test_data["rank"])
        assert snpa.H.shape == (test_data["rank"], test_data["X"].shape[1])
        assert len(snpa.col_indices_ls) == test_data["rank"]

        # check non-negativity
        assert np.min(snpa.W) >= 0
        assert np.min(snpa.H) >= 0



import warnings
import numpy as np
from numpy.typing import NDArray

from ._utils import simplex_proj, simplex_proj_using_axis, unimodal_reg_l2


class FPGM:
    """
    An implementation of Non-negative Least Squares solver using Fast Projected Gradient
    Method.

    Solves for :math:`H` given :math:`X` and :math:`W` by minimizing the following
    objective function with other optional constraints on :math:`H`

    .. math::
        ||X - WH||_{F}^2

    where :math:`||\\cdot||_F` is the Frobenius norm.

    Parameters
    ----------
    constraint_kind : integer-like {0, 1, 2, 3}, default=0
        The following constraints are applied based on the integer value specified

        * If ``0``: Only :math:`H \\geq 0`.
        * If ``1``: Closure constraint :math:`H^T e ≤ e`.
        * If ``2``: Closure constraint :math:`H e = e`.
        * If ``3``: Closure constraint :math:`H^T e = e`.
    iter_max : int, default=500
        Maximum number of iterations.
    tol : float, default=1e-6
        Tolerance for convergence. Must be in the interval :math:`(0, 1)`.

            Convergence is reached when:

            .. math::

                {||H^{i} - H^{i-1}||_{F} \\over ||H^{1} - H^{0}||_{F}} \\leq
                \\textrm{tol}

            where :math:`H^{i}` is :math:`H` after iteration :math:`i` and :math:`H^{0}` is
            initial :math:`H`.

    References
    ----------
    .. [1] Gillis, N. (2020). Nonnegative matrix factorization. Society for Industrial
           and Applied Mathematics.
    .. [2] Lin, Chih-Jen. "Projected gradient methods for nonnegative matrix
           factorization." Neural computation 19.10 (2007): 2756-2779.

    Examples
    --------
    >>> from mcrnmf.nnls import FPGM
    >>> from mcrnmf import SNPA
    >>> from mcrnmf.datasets import load_rxn_spectra
    >>>
    >>> # load the dataset available in the package
    >>> X, wv, time = load_rxn_spectra()
    >>> # Get an estimate for W from SNPA
    >>> snpa = SNPA(rank=4)
    >>> snpa.fit(X=X)
    >>> W = snpa.W
    >>>
    >>> # create an instance of FPGM
    >>> fpgm = FPGM(constraint_kind=1, iter_max=400, tol=1e-4)
    >>> # solve for H
    >>> H, is_converged = fpgm.solve(X=X, W=W, unimodal_H=True)

    """

    def __init__(self, constraint_kind=0, iter_max=500, tol=1e-6):
        # validate types
        if not isinstance(constraint_kind, (int, np.integer)):
            raise TypeError(
                f"`constraint_kind` must be of type {int}, current type is "
                f"{type(constraint_kind)}"
            )
        if not isinstance(iter_max, (int, np.integer)):
            raise TypeError(
                f"`iter_max` must be of type {int}, current type is {type(iter_max)}"
            )
        if not isinstance(tol, float):
            raise TypeError(
                f"`tol` must be of type {float}, current type is {type(tol)}"
            )

        # validate values
        if constraint_kind not in [0, 1, 2, 3]:
            raise ValueError("`constraint_kind` must be in 0, 1, 2, or 3.")
        if iter_max <= 0:
            raise ValueError("`iter_max` must be > 0")
        if not (0 < tol < 1):
            raise ValueError("`tol` must be in the open interval (0, 1)")

        self.constraint_kind_ = constraint_kind
        self.iter_max_ = iter_max
        self.tol_ = tol

    def solve(
        self,
        X: NDArray[np.float64],
        W: NDArray[np.float64],
        H: NDArray[np.float64] | None = None,
        known_H: NDArray[np.float64] | None = None,
        unimodal_H: bool | list[bool] = False,
    ):
        """
        Solves for :math:`H` given :math:`X` and :math:`W`.

        Parameters
        ----------
        X : ndarray of shape (m, n)
            Data array.

        W : ndarray of shape (m, k)
            Coefficient array.

        H : ndarray of shape (k, n) or None, defualt=None
            Initial guess for the solution array. If ``None``, an initial guess is
            generated using least squares solution of :math:`WH \\approx X`.

        known_H : ndarray of shape (rank, n_samples), default=None
            Array containing known values of :math:`H`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`H` which do not
              correspond ``np.nan`` entries in ``known_H``.

        unimodal_H : bool or list of bool, default=False
            Applies unimodality constraint to rows of H.

                * If ``bool`` and ``True`` applies unimodality constraint to all rows
                  of `H`
                * If ``list[bool]``:

                    * Length of list must be equal to number of columns in `W`.
                    * Applies unimodality constraint only those rows of `H` which are
                      marked as ``True``. Example: [True, False, True] applies
                      unimodality constraint to rows 1 and 3 of `H`.

        Returns
        -------
        H : ndarray of shape (k, n)
            Solution array.

        is_converged : bool
            Whether algorithm converged within ``iter_max`` iterations.

            * ``True`` if convergence was reached based on ``tol`` criterion
            * ``False`` if maximum iterations were reached without convergence

        """

        X, W, H = _setup_input_args(X=X, W=W, H=H)

        # check known_H
        if not ((known_H is None) or (isinstance(known_H, np.ndarray))):
            raise TypeError("`known_H` must be None or a numpy array.")
        elif isinstance(known_H, np.ndarray):
            if known_H.shape[0] != W.shape[1]:
                raise ValueError(
                    "Shape mismatch: number of columns of `known_H` not equal to "
                    "number of columns of `W`"
                )
        if known_H is not None:
            mask_known_H = np.isfinite(known_H)
            tot_fvals_knw_H = np.sum(mask_known_H)
            if tot_fvals_knw_H == 0:
                warnings.warn(
                    "`known_H` contains no finite values, setting it to None."
                )
                known_H = None
            elif tot_fvals_knw_H == known_H.size:
                raise ValueError(
                    "All values of `H` are specified as known, can't solve."
                )
            else:
                # finite elems are non-neg
                if known_H[mask_known_H].min() < 0:
                    raise ValueError("All finite elements of `known_H` must be >= 0.")
        # using if statement instead of else with previous because known_H is set to
        # None if none of the values are finite.
        if known_H is None:
            mask_known_H = None
        # check unimodal arg
        if not isinstance(unimodal_H, (bool, list)):
            raise TypeError(f"`unimodal_H` must be of type bool or list of bool")
        elif isinstance(unimodal_H, list):
            if not all(isinstance(elem, bool) for elem in unimodal_H):
                raise TypeError(
                    f"If `unimodal_H` is a list, it must contain only boolean elements"
                )
        if isinstance(unimodal_H, list):
            if len(unimodal_H) != W.shape[1]:
                raise ValueError(
                    "If `unimodal_H` is specified as list, then the length of list "
                    "must be equal to number of columns of W"
                )
        else:
            unimodal_H = [unimodal_H] * W.shape[1]

        # precompute WtW, and WtX since it doesn't change in the loop
        WtW = W.T @ W
        WtX = W.T @ X

        # hyperparameters for optimization
        l = np.linalg.norm(WtW, 2)  # lipschitz constant
        theta_prev = 0.05  # first value of theta, heuristic

        # pre-allocate memory for matrices computed or used repeatedly in the loop
        Y = H.copy()  # look ahead variable
        Y_grad = np.zeros(H.shape, dtype=H.dtype)
        H_diff = np.zeros(H.shape, dtype=H.dtype)
        H_prev = np.zeros(H.shape, dtype=H.dtype)

        i = 0
        is_converged = False
        while i < self.iter_max_:
            np.copyto(H_prev, H)
            # compute the gradient using look ahead point
            np.matmul(WtW, Y, out=Y_grad)
            np.subtract(Y_grad, WtX, out=Y_grad)
            np.divide(Y_grad, l, out=Y_grad)
            # gradient step: H = Y - ((W.T @ W @ H) - (W.T @ X))) / l
            np.subtract(Y, Y_grad, out=H)

            if mask_known_H is not None:
                H[mask_known_H] = known_H[mask_known_H]
            self._apply_unimodal_H_rows(H=H, unimodal_H=unimodal_H)
            self._apply_projection(A=H, constraint_kind=self.constraint_kind_)

            np.subtract(H, H_prev, out=H_diff)
            norm_H_diff = np.linalg.norm(H_diff, "fro")
            if i == 0:
                norm_H_diff_0 = norm_H_diff

            # convergence check
            if norm_H_diff <= (self.tol_ * norm_H_diff_0):
                is_converged = True
                break

            # do the momentum update only if convergence criteria is not satisfied
            theta_prev_sq = theta_prev**2
            # theta and beta computation based on scheme 2.2.19 in Introductory lectures
            # on convex optimization by Nesterov
            theta = (np.sqrt(theta_prev_sq**2 + 4 * theta_prev_sq) - theta_prev_sq) / 2
            beta = theta_prev * (1 - theta_prev) / (theta_prev_sq + theta)
            # momentum step
            np.multiply(H_diff, beta, out=H_diff)
            np.add(H, H_diff, out=Y)

            theta_prev = theta
            i += 1

        return H, is_converged

    def _apply_unimodal_H_rows(self, H, unimodal_H):
        """
        Apply unimodal constraints to rows of H in-place.

        Note: This is private method intended for internal use only.

        Parameters
        ----------
        H : NDArray[np.float64]
            The current value of H
        unimodal_H : list of bool
            A list with boolean values specifying to which rows of H unimodal
            constraints should be applied.
        """
        nrows, ncols = H.shape
        w = np.ones(ncols)
        for row in range(nrows):
            if unimodal_H[row] is True:
                H[row, :] = unimodal_reg_l2(y=H[row, :], w=w)

    def _apply_projection(self, A, constraint_kind):
        """
        Perform in-place projection of A based on the constraint.

        Note: This is private method intended for internal use only.

        Parameters
        ----------
        A : NDArray[np.float64]
            H matrix to be projected.

        constraint_kind: int | None
            The kind of constraint to use during projection.

            If constraint_kind=0, then A >= 0

            If constraint_kind=1 then A^Te <= e and A >= 0

            If constraint_kind=2 then Ae = e and A >= 0

            If constraint_kind=3 then A^Te = e and A >= 0
        """
        if constraint_kind == 0:
            # performs projection such that H >= 0
            np.maximum(A, 0, out=A)
        elif constraint_kind == 1:
            # performs projection such that H^T*e <= e and H >= 0
            simplex_proj(A)
        elif constraint_kind == 2:
            # performs projection such that H*e = e and H >= 0
            A[:] = simplex_proj_using_axis(A, is_col=False)
        elif constraint_kind == 3:
            # performs projection such that H^T*e = e and H >= 0
            A[:] = simplex_proj_using_axis(A, is_col=True)


def _setup_input_args(X, W, H):
    """
    Validate the shapes of matrices used in NNLS and generates an initial guess for H
    if it is None, else validates it

    Note: This is a private method intended for internal use only.

    Parameters
    ----------
    X : ndarray
        Input data matrix of shape (m, n).

    W : ndarray
        Left factor matrix of shape (m, r).

    H : ndarray | None, optional
        Initial matrix of shape (r, n) to be updated. Default is None.

        If None, then an initial guess of `H` is generated.

    """
    # check type
    if not isinstance(X, np.ndarray):
        raise TypeError("`X` must be a numpy array.")
    if not isinstance(W, np.ndarray):
        raise TypeError("`W` must be a numpy array.")

    is_H_guessed = False
    if H is None:
        H = _initialize_H(X=X, W=W)
        is_H_guessed = True
    else:
        if not isinstance(H, np.ndarray):
            raise TypeError("`H` must be a numpy array or None.")
    # check shape matches
    m, n = X.shape
    rank = W.shape[1]

    if W.shape[0] != m:
        raise ValueError(
            f"Shape mismatch: rows of `W`: {W.shape[0]} not equal to rows of X: {m}"
        )
    if not is_H_guessed:
        if H.shape[0] != rank:
            raise ValueError(
                f"Shape mismatch: rows of `H`: {H.shape[0]} not equal to columns of W: "
                f"{rank}"
            )
        if H.shape[1] != n:
            raise ValueError(
                f"Shape mismatch: columns of `H`: {H.shape[1]} not equal to columns of "
                f"X: {n}."
            )

    if not np.all(H >= 0):
        raise ValueError("All elements of `H` must be >= 0.")

    # ensure all arrays are of type float64 and C-contiguous
    X = np.ascontiguousarray(X.copy(), dtype=np.float64)
    W = np.ascontiguousarray(W.copy(), dtype=np.float64)
    H = np.ascontiguousarray(H.copy(), dtype=np.float64)

    return X, W, H


def _initialize_H(X, W):
    """
    Generated initial guess for H in the NNLS problem.

        min_{H>=0} ||X - WH||_F^2

    Parameters
    ----------
    X : np.ndarray
        Input data matrix of shape (m, n)

    W : np.ndarray
        Matrix of shape (m, k)
    """

    if np.linalg.cond(W) > 1e6:
        warnings.warn(
            "`W` has very high condition number. Initial guess of H generated might be "
            "poor."
        )
    H = np.linalg.lstsq(W, X, rcond=None)[0]
    np.maximum(H, 0, out=H)
    alpha = np.sum(H * (W.T @ X)) / np.sum((W.T @ W) * (H @ H.T))
    H *= alpha  # scale H

    # replaces rows which sum to zero with small random values
    zero_sum_rows = np.where(np.sum(H, axis=1) == 0)[0]
    if len(zero_sum_rows) > 0:
        H[zero_sum_rows, :] = (
            0.001 * np.max(H) * np.random.rand(len(zero_sum_rows), H.shape[1])
        )

    return H



import numpy as np
import numpy.typing as npt
from numba import njit


@njit
def prefix_isotonic_regression(
    y: npt.NDArray[np.float64], w: npt.NDArray[np.float64], is_increasing: bool
) -> npt.NDArray[np.float64]:
    """
    Perform prefix isotonic regression on a given sequence with weights.

    This function computes the prefix isotonic regression of a given sequence `y`
    with corresponding weights `w`. The regression can be either increasing or
    decreasing based on the `is_increasing` flag. The function returns an array
    of errors associated with the isotonic regression at each prefix of the
    sequence.

    Parameters
    ----------
    y : NDArray[np.floating]
        The input 1D array of values to be isotonic-regressed.
    w : NDArray[np.floating]
        The 1D array of weights corresponding to the values in `y`.
    is_increasing : bool
        A flag indicating the direction of the isotonic regression.

            - If True, the regression is performed in an increasing manner.

            - If False, the regression is performed in a decreasing manner.

    Returns
    -------
    np.ndarray
        An array of errors for each prefix of the sequence after performing
        isotonic regression.
    """
    n = y.size
    if is_increasing:
        mean = y.copy()
        sum_wy2 = y**2 * w**2
        sum_wy = w * y
        sum_w = w.copy()
    else:
        mean = y[::-1].copy()
        sum_wy = w[::-1] * y[::-1]
        sum_wy2 = (w[::-1] ** 2) * (y[::-1] ** 2)
        sum_w = w[::-1].copy()

    error = np.zeros(n)
    lvl_error = np.zeros(n)
    vec_back = 0
    for j in range(1, n):
        error[j] = error[j - 1]
        while (vec_back >= 0) and (mean[j] <= mean[vec_back]):
            sum_wy[j] += sum_wy[vec_back]
            sum_wy2[j] += sum_wy2[vec_back]
            sum_w[j] += sum_w[vec_back]
            mean[j] = sum_wy[j] / sum_w[j]
            lvl_error[j] = sum_wy2[j] - (sum_wy[j] ** 2 / sum_w[j])
            error[j] -= lvl_error[vec_back]
            vec_back -= 1
        vec_back += 1
        sum_wy[vec_back] = sum_wy[j]
        sum_wy2[vec_back] = sum_wy2[j]
        sum_w[vec_back] = sum_w[j]
        mean[vec_back] = mean[j]
        lvl_error[vec_back] = lvl_error[j]
        error[j] += lvl_error[vec_back]

    return error


def simplex_proj(Y: npt.NDArray[np.float64]):
    """
    In-place projection of matrix Y onto the simplex.

    Solves:

        min_{x>=0} ||x - y(:, j)||_2  for all j

            s.t:   x^T e <= e

    If Y is a 2D array of shape [d, n] then the n points in d

    Parameters
    ----------
    Y : ndarray
        Matrix or vector to be projected.

        If Y is a 2D array of shape `(d, n)`, then the `n` columns are project to a
        `d` dimensional simplex

        If Y is a 1D array of shape `(n,)` then each element of the array is
        projected to a 1-dimensional simplex, i.e., a line segment.

    """
    np.maximum(Y, 0, out=Y)
    if Y.ndim == 1:
        # This case would arise only while using rank = 1 during factorization.
        np.minimum(Y, 1, out=Y)
    else:
        col_idx_arr = np.where(np.sum(Y, axis=0) > 1)[0]
        # although _simplex_col_proj does the projection in-place we need the
        # reassignment because Y[:, col_idx_arr] slicing creates a copy. Hence the
        # in-place operation is done on the copy.
        if col_idx_arr.size > 0:
            Y[:, col_idx_arr] = simplex_proj_using_axis(Y[:, col_idx_arr], is_col=True)


def simplex_proj_using_axis(B: npt.NDArray[np.float64], is_col: bool):
    """
    Performs in-place projection of matrix B's columns or rows to simplex and
    non-negative orthant.

    Parameters
    ----------
    B : NDArray[np.float64]
        Matrix to be projected.

    is_col : bool
        If True, project columns; if False, project rows.

    Returns
    -------
    NDArray[np.float64]
        Projected matrix B

    """
    if is_col:
        d, n = B.shape
        B_sorted = np.sort(B, axis=0)[::-1, :]
        B_temp = (np.cumsum(B_sorted, axis=0) - 1) / (
            np.arange(1, d + 1)[:, np.newaxis]
        )
        rho = np.sum(B_sorted > B_temp, axis=0)
        lambdaa = B_temp[rho - 1, np.arange(n)]
        np.maximum(B - lambdaa, 0, out=B)
    else:
        n, d = B.shape
        B_sorted = np.sort(B, axis=1)[:, ::-1]
        B_temp = (np.cumsum(B_sorted, axis=1) - 1) / np.arange(1, d + 1)
        rho = np.sum(B_sorted > B_temp, axis=1)
        lambdaa = B_temp[np.arange(n), rho - 1]
        np.maximum(B - lambdaa[:, np.newaxis], 0, out=B)

    return B


@njit
def unimodal_reg_l2(
    y: npt.NDArray[np.float64], w: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """
    Perform unimodal regression with L2 loss on a given sequence `y`.

    This function computes the unimodal regression of a given sequence `y`
    with corresponding weights `w`. It finds the mode of the sequence and
    performs isotonic regression on both the ascending and descending parts.

    Based on the cpp implementation in UniIsoRegression package (see [1]_)

    Parameters
    ----------
    y : np.ndarray
        The input 1D array of values to be unimodal-regressed.
    w : np.ndarray
        The 1D array of weights corresponding to the values in `y`.

    Returns
    -------
    np.ndarray
        The unimodal-regressed sequence.

    Notes
    -----
    The function uses a helper function `prefix_isotonic_regression` to
    perform isotonic regression on prefixes of the sequence.

    References
    ----------
    .. [1] https://github.com/xzp1995/UniIsoRegression

    """

    # compute errors for ascending and descending isotonic regression
    error_asc = prefix_isotonic_regression(y=y, w=w, is_increasing=True)
    error_des = prefix_isotonic_regression(y=y, w=w, is_increasing=False)
    # find the mode by minimizing total error
    n = y.size
    total_error = np.zeros(n + 1)
    total_error[0] = error_des[-1]
    total_error[-1] = error_asc[-1]
    total_error[1:-1] = error_asc[:-1] + np.flip(error_des[1:])
    # there could be multiple modes with same minima. In those cases it picks the first
    # minima idx as mode
    mode_idx = np.argmin(total_error)

    # find the unimodal fit
    left = np.arange(n, dtype=np.int64)
    right = np.arange(n, dtype=np.int64)
    mean = np.zeros(n)
    sum_wy = np.zeros(n)
    sum_w = np.zeros(n)
    yfit = np.zeros(n)

    # perform regression on ascending part
    if mode_idx != 0:
        mean[:mode_idx] = y[:mode_idx]
        sum_wy[:mode_idx] = y[:mode_idx] * w[:mode_idx]
        sum_w[:mode_idx] = w[:mode_idx]
        vec_back = 0
        for j in range(1, mode_idx):
            while (vec_back >= 0) and (mean[j] <= mean[vec_back]):
                left[j] = left[vec_back]
                sum_wy[j] += sum_wy[vec_back]
                sum_w[j] += sum_w[vec_back]
                mean[j] = sum_wy[j] / sum_w[j]
                vec_back -= 1
            vec_back += 1
            left[vec_back] = left[j]
            right[vec_back] = right[j]
            sum_wy[vec_back] = sum_wy[j]
            sum_w[vec_back] = sum_w[j]
            mean[vec_back] = mean[j]
        for k in range(vec_back + 1):
            yfit[left[k] : right[k] + 1] = mean[k]

    # perform regression on descending part
    if mode_idx != n:
        mean[: n - mode_idx] = np.flip(y[mode_idx:])
        left[: n - mode_idx] = np.arange(n - mode_idx, dtype=np.int64)
        right[: n - mode_idx] = np.arange(n - mode_idx, dtype=np.int64)
        sum_w[: n - mode_idx] = np.flip(w[mode_idx:])
        sum_wy[: n - mode_idx] = np.flip(w[mode_idx:] * y[mode_idx:])

        vec_back = 0
        for j in range(1, n - mode_idx):
            while (vec_back >= 0) and (mean[j] <= mean[vec_back]):
                left[j] = left[vec_back]
                sum_wy[j] += sum_wy[vec_back]
                sum_w[j] += sum_w[vec_back]
                mean[j] = sum_wy[j] / sum_w[j]
                vec_back -= 1
            vec_back += 1
            left[vec_back] = left[j]
            right[vec_back] = right[j]
            sum_wy[vec_back] = sum_wy[j]
            sum_w[vec_back] = sum_w[j]
            mean[vec_back] = mean[j]
        for k in range(vec_back + 1):
            yfit[n - right[k] - 1 : n - left[k]] = mean[k]

    # mode index w.r.t to input array since `total_error` has n + 1 elements.
    mode_idx -= 1

    return yfit



"""
MCR-NMF: Multivariate Curve Resolution with Non-negative Matrix Factorization

A Python package implementing various non-negative matrix factorization (NMF) algorithms
for multivariate curve resolution (MCR) applications.
"""

# To get sub-modules
from .models import FroALS, FroFPGM, MinVol, SNPA
from .nnls import FPGM

__version__ = "0.1.0"

__all__ = ["FroALS", "FroFPGM", "MinVol", "SNPA", "FPGM"]



from abc import ABC, abstractmethod
import warnings

import numpy as np
from numpy.typing import NDArray

from ._utils import unimodal_reg_l2, simplex_proj, simplex_proj_using_axis

_EPSILON = np.finfo(np.float64).eps


class _BaseNMF(ABC):
    """
    Base class for Non-negative Matrix Factorization (NMF) algorithms.

    This abstract class provides the foundation for various NMF implementations.
    It handles common functionality such as convergence checking, error
    computation, and projection operations for enforcing constraints.

    NMF decomposes a non-negative matrix X into two lower-rank non-negative
    matrices: X ≈ W * H, where W and H are the factor matrices.

    Parameters
    ----------
    rank : int
        The rank of the factorization (number of components).
    constraint_kind : int, optional
        The type of constraint to apply (default is 0):
        - 0: Default, Non-negativity constraints only (W ≥ 0, H ≥ 0)
        - 1: Simplex constraint (H^T e ≤ e)
        - 2: Row simplex constraint (H e = e)
        - 3: Column simplex constraint on W (W^T e = e)
        - 4: Column simplex constraint on H (H^T e = e)
    unimodal : dict or None, default=None
        Specifies unimodality constraints for W and H matrices.
        If None, no unimodality constraints are applied.

        If dict, Format: {'W': [bool list or bool], 'H': [bool list or bool]}
        * Must contain at least one key ('W' or 'H')
        * No other keys besides 'W' and 'H' are allowed
        * For 'W': Controls unimodality of columns in W
        * For 'H': Controls unimodality of rows in H

        Each value can be:

        * A boolean: applies to all components
        * A list of booleans: selectively applies to specific components

        Examples:

        * {'W': True, 'H': True} - All :math:`W` and :math:`H` components are unimodal
        * {'H': [True, False, True]} - Only components 0 and 2 of :math:`H` have
          unimodal behavior
        * {'W': [False, True, False]} - Only component 1 of :math:`W` has unimodal
          behavior
    iter_max : int, optional
        Maximum number of outer iterations for the algorithm (default is 500).
    tol : float, optional
        Convergence tolerance for the relative loss (default is 1e-4).

    Methods
    -------
    fit(X, Wi, Hi, known_W=None, known_H=None)
        Fit the NMF model to the provided data matrix.

    Raises
    ------
    TypeError
        If input parameters are not of the correct type.
    ValueError
        If input parameters have invalid values.

    Notes
    -----
    This is an abstract base class. Concrete implementations should override
    the `fit`, `_update_H`, and `_update_W` methods.
    """

    def __init__(
        self,
        rank: int,
        constraint_kind: int = 0,
        unimodal: dict | None = None,
        iter_max: int = 500,
        tol: float = 1e-4,
    ):
        # validate type
        if not isinstance(rank, (int, np.integer)):
            raise TypeError(
                f"`rank` must be of type {int}, currrent type is {type(rank)}"
            )
        if not isinstance(constraint_kind, (int, np.integer)):
            raise TypeError(
                f"`constraint_kind` must be of type {int}, current type is "
                f"{type(constraint_kind)}"
            )
        if unimodal is not None:
            if not isinstance(unimodal, dict):
                raise TypeError(f"`unimodal` must be None or a dictionary")
        if not isinstance(iter_max, (int, np.integer)):
            raise TypeError(
                f"`iter_max` must be of type {int}, current type is {type(iter_max)}"
            )
        if not isinstance(tol, float):
            raise TypeError(
                f"`tol` must be of type {float}, current type is {type(tol)}"
            )

        # validate values
        if rank <= 0:
            raise ValueError("`rank` must be > 0")
        if constraint_kind not in [0, 1, 2, 3, 4]:
            raise ValueError("`constraint_kind` must be 0, 1, 2, 3, or 4")
        # default unimodal constraints if None
        unimodal_W = [False] * rank
        unimodal_H = [False] * rank
        if unimodal is not None:
            # empty dictionary check
            if not unimodal:
                raise ValueError(
                    "If not None, unimodal dictionary cannot be empty; must contain "
                    "'W' and/or 'H' keys."
                )
            # check for required keys
            valid_keys = ["W", "H"]
            provided_keys = list(unimodal.keys())
            for provided_key in provided_keys:
                if provided_key not in valid_keys:
                    raise KeyError(
                        f"`unimodal` dictionary can contain only keys in: {valid_keys} "
                        f"got {provided_key}."
                    )
            if "W" in unimodal:
                unimodal_W = self._process_unimodal_arg(unimodal["W"], rank)
            if "H" in unimodal:
                unimodal_H = self._process_unimodal_arg(unimodal["H"], rank)
        if iter_max < 10:
            raise ValueError("`iter_max` must be >= 10")
        if not (0 < tol < 1):
            raise ValueError("`tol` must be in the open interval (0, 1)")

        self.rank_ = rank
        self.constraint_kind_ = constraint_kind
        self.unimodal_W_ = unimodal_W
        self.unimodal_H_ = unimodal_H
        self.iter_max_ = iter_max
        self.tol_ = tol

        # initialize attributes
        self._W = None
        self._H = None
        self._rel_loss_ls = None
        self._is_converged = None

    @property
    def W(self):
        """
        The basis matrix :math:`W` obtained after fitting the model.

        Returns
        -------
        ndarray of shape (n_features, rank)
            The obtained :math:`W` after fitting the model.
        """
        if self._W is None:
            raise AttributeError(
                "`W` is not available. Model has not been fitted yet. Call 'fit' first."
            )
        return self._W

    @property
    def H(self):
        """
        The coefficient matrix :math:`H` obtained after fitting the model.

        Returns
        -------
        ndarray of shape (rank, n_samples)
            The obtained :math:`H` after fitting the model.
        """
        if self._H is None:
            raise AttributeError(
                "`H` is not available. Model has not been fitted yet. Call 'fit' first."
            )
        return self._H

    @property
    def rel_loss_ls(self):
        """
        List of relative loss values from each iteration during model fitting.

        It is defined as:

        .. math::

            \\dfrac{\\sqrt{f_{\\textrm{obj}}^{i}}}{||X||_F}\\ \\textrm{,}

        where:

        * :math:`||\\cdot||_F` denotes the Frobenius norm
        * :math:`X` is the original data matrix
        * :math:`f_{\\textrm{obj}}^{i}` is the value of objective function after
          iteration :math:`i`

        Returns
        -------
        list of float
            Relative loss value from each iteration.
        """
        if self._rel_loss_ls is None:
            raise AttributeError(
                "`rel_loss_ls` is not available. Model has not been fitted yet. Call "
                "`fit` first."
            )
        return self._rel_loss_ls

    @property
    def is_converged(self):
        """
        The convergence status.

        Returns
        -------
        bool
            Whether the algorithm converged within ``iter_max`` iterations.

            * ``True`` if convergence was reached based on ``tol`` criterion
            * ``False`` if maximum iterations were reached without convergence
        """
        if self._is_converged is None:
            raise AttributeError(
                "`is_converged` is not available. Model has not been fitted yet. Call "
                "`fit` first."
            )
        return self._is_converged

    @abstractmethod
    def fit(
        self,
        X: NDArray[np.float64],
        Wi: NDArray[np.float64],
        Hi: NDArray[np.float64],
    ) -> None:
        """
        Fit the NMF model to the provided data matrix.

        This method must be implemented by derived classes. Different NMF
        algorithms implement this method differently.
        """
        self._W = Wi.copy()
        self._H = Hi.copy()
        self._rel_loss_ls = []
        self._is_converged = False

    @abstractmethod
    def _update_H(self) -> None:
        """
        Update the H matrix based on the current W matrix.

        This method must be implemented by derived classes. Different NMF
        algorithms implement this method differently.
        """
        pass

    @abstractmethod
    def _update_W(self) -> None:
        """
        Update the W matrix based on the current H matrix.

        This method must be implemented by derived classes. Different NMF
        algorithms implement this method differently.
        """
        pass

    def _process_unimodal_arg(self, constraint: bool | list[bool], rank: int):
        """
        Process unimodal constraint for either H or W key into a boolean list of length
        rank.
        """
        if isinstance(constraint, bool):
            return [constraint] * rank
        elif isinstance(constraint, list):
            if len(constraint) != rank:
                raise ValueError(f"Unimodal constraint list must have length {rank}.")
            if not all(isinstance(c, bool) for c in constraint):
                raise TypeError("Unimodal constraint list must contain only booleans.")
            return constraint
        else:
            raise TypeError("Unimodal constraint must be bool or list of booleans")

    def _apply_unimodal_H_rows(self) -> None:
        """Apply unimodal constraints to rows of H."""
        nrows, ncols = self._H.shape
        w = np.ones(ncols)
        for row in range(nrows):
            if self.unimodal_H_[row] is True:
                self._H[row, :] = unimodal_reg_l2(y=self._H[row, :], w=w)

    def _apply_unimodal_W_cols(self) -> None:
        """Apply unimodal constraints to columns of W."""
        nrows, ncols = self._W.shape
        w = np.ones(nrows)
        for col in range(ncols):
            if self.unimodal_W_[col]:
                self._W[:, col] = unimodal_reg_l2(y=self._W[:, col], w=w)

    def _check_convergence(self, iter: int) -> bool:
        """
        Check if the algorithm has converged.

        Parameters
        ----------
        iter : int
            Current iteration number.

        Returns
        -------
        bool
            True if the algorithm has converged, False otherwise.
        """
        if iter < 10:
            is_converged = False
        else:
            is_converged = (
                np.abs(self._rel_loss_ls[iter - 10] - self._rel_loss_ls[iter])
                / np.maximum(_EPSILON, self._rel_loss_ls[iter])
                < self.tol_
            )

        return is_converged

    def _fro_err_sq_after_W_update(
        self, norm_X: float, HHt: NDArray[np.float64], XHt: NDArray[np.float64]
    ) -> float:
        """
        Calculate the squared Frobenius error after W update.

        Computes ||X - W*H||_F^2 efficiently using precomputed matrices.

        Parameters
        ----------
        norm_X : float
            Frobenius norm of X.
        HHt : ndarray
            Precomputed H * H^T matrix.
        XHt : ndarray
            Precomputed X * H^T matrix.

        Returns
        -------
        float
            Squared Frobenius error.
        """
        fro_error_sq = np.maximum(
            _EPSILON,
            norm_X**2 - 2 * np.sum(XHt * self._W) + np.sum(HHt * (self._W.T @ self._W)),
        )

        return fro_error_sq

    def _fro_err_sq_after_H_update(
        self, norm_X: float, WtW: NDArray[np.float64], WtX: NDArray[np.float64]
    ) -> float:
        """
        Calculate the squared Frobenius error after H update.

        Computes ||X - W*H||_F^2 efficiently using precomputed matrices.

        Parameters
        ----------
        norm_X : float
            Frobenius norm of X.
        WtW : ndarray
            Precomputed W^T * W matrix.
        WtX : ndarray
            Precomputed W^T * X matrix.

        Returns
        -------
        float
            Squared Frobenius error.
        """
        fro_error_sq = np.maximum(
            _EPSILON,
            norm_X**2 - 2 * np.sum(WtX * self._H) + np.sum(WtW * (self._H @ self._H.T)),
        )

        return fro_error_sq

    def _project_H(self):
        """
        Project the coefficient matrix H onto the constraint set.

        Applies different projections to the H matrix based on the specified
        constraint type. The projection ensures that H satisfies the constraints
        required for the non-negative matrix factorization problem.

        The following constraints are implemented:
        - constraint_kind = 0: Non-negativity constraint (H >= 0)
        - constraint_kind = 1: Simplex constraint (H^T e <= e)
        - constraint_kind = 2: Row simplex constraint (H e = e)
        - constraint_kind = 3: Non-negativity constraint (H >= 0, same as 0)
        - constraint_kind = 4: Column simplex constraint (H^T e = e)
        """
        if self.constraint_kind_ in [0, 3]:
            self._H[:] = np.maximum(self._H, 0.0)
        elif self.constraint_kind_ == 1:
            simplex_proj(self._H)
        elif self.constraint_kind_ == 2:
            self._H[:] = simplex_proj_using_axis(B=self._H, is_col=False)
        elif self.constraint_kind_ == 4:
            self._H[:] = simplex_proj_using_axis(B=self._H, is_col=True)

    def _project_W(self):
        """
        Project the basis matrix W onto the constraint set.

        Applies different projections to the W matrix based on the specified
        constraint type. The projection ensures that W satisfies the constraints
        required for the non-negative matrix factorization problem.

        The following constraints are implemented:
        - constraint_kind = 3: Column simplex constraint (W >= 0, W^T e = e)
        - Any other value: Non-negativity constraint (W >= 0)
        """
        if self.constraint_kind_ == 3:
            self._W[:] = simplex_proj_using_axis(B=self._W, is_col=True)
        else:
            self._W[:] = np.maximum(self._W, 0.0)

    def _validate_fit_args(
        self,
        X: NDArray,
        Wi: NDArray,
        Hi: NDArray,
        known_W: NDArray | None,
        known_H: NDArray | None,
    ):
        """
        Validate input arguments for matrix factorization.

        This function checks the dimensions and values of the input matrices
        to ensure they meet the requirements for non-negative matrix factorization.

        Parameters
        ----------
        X : ndarray
            The input data matrix to be factorized.
        Wi : ndarray
            Initial guess for the W matrix.
        Hi : ndarray
            Initial guess for the H matrix.
        known_W : ndarray or None
            Matrix with known values of W.
        known_H : ndarray or None
            Matrix with known values of H.

        Raises
        ------
        TypeError
            If any input matrix is not a numpy array or has incorrect dtype.
        ValueError
            If any dimension compatibility check fails or if any matrix
            contains negative values.

        Warns
        -----
        UserWarning
            If known_W or known_H contains no finite values.
        """
        # check type
        if not isinstance(X, np.ndarray):
            raise TypeError("`X` must be a numpy array.")
        if not isinstance(Wi, np.ndarray):
            raise TypeError("`Wi` must be a numpy array.")
        if not isinstance(Hi, np.ndarray):
            raise TypeError("`Hi` must be a numpy array.")
        if not ((known_W is None) or (isinstance(known_W, np.ndarray))):
            raise TypeError("`known_W` must be None or a numpy array.")
        if not ((known_H is None) or (isinstance(known_H, np.ndarray))):
            raise TypeError("`known_H` must be None or a numpy array.")

        # check shape matches
        rows_X, cols_X = X.shape
        if Wi.shape[0] != rows_X:
            raise ValueError(
                "Shape mismatch: number of rows of `Wi` not equal to number of rows of "
                "`X`"
            )
        if Wi.shape[1] != self.rank_:
            raise ValueError(
                "Shape mismatch: number of columns of `Wi` not equal to `rank`"
            )
        if Hi.shape[1] != cols_X:
            raise ValueError(
                "Shape mismatch: number of columns of `Hi` not equal to number of "
                "columns of `X`"
            )
        if Hi.shape[0] != self.rank_:
            raise ValueError(
                f"Shape mismatch: number of rows of `Hi` not equal to `rank`"
            )
        # check shape matches between known_W, known_H, initial_W, initial_H
        if (known_W is not None) and (known_W.shape != Wi.shape):
            raise ValueError(f"Shape mismatch: `known_W` shape not equal to `Wi` shape")
        if (known_H is not None) and (known_H.shape != Hi.shape):
            raise ValueError(f"Shape mismatch: `known_H` shape not equal to `Hi` shape")

        # check dtype is np.float64
        if X.dtype != np.float64:
            raise TypeError(
                f"The dtype of `X` elements should be {np.float64}, current dtype is "
                f"{X.dtype}"
            )
        if Wi.dtype != np.float64:
            raise TypeError(
                f"The dtype of `Wi` elements should be {np.float64}, current dtype is "
                f"{Wi.dtype}"
            )
        if Hi.dtype != np.float64:
            raise TypeError(
                f"The dtype of `Hi` elements should be {np.float64}, current dtype is "
                f"{Hi.dtype}"
            )
        # check if non-negative
        if X.min() < 0:
            raise ValueError("All elements of `X` must be >= 0")
        if Wi.min() < 0:
            raise ValueError("All elements of `Wi` must be >= 0")
        if Hi.min() < 0:
            raise ValueError("All elements of `Hi` must be >= 0")
        # check the specific validation arguments
        if known_W is not None:
            mask_fvals_knw_W = np.isfinite(known_W)
            tot_fvals_knw_W = np.sum(mask_fvals_knw_W)
            if tot_fvals_knw_W == 0:  # if all elements are NaN
                warnings.warn(
                    "`known_W` contains no finite values, setting it to None."
                )
                known_W = None
            elif tot_fvals_knw_W == known_W.size:  # if all elements are finite
                raise ValueError(
                    "All values of `W` are known, can't perform NMF. Try `NNLS solver` "
                    "to compute H."
                )
            else:
                # finite elems are non-neg
                if known_W[mask_fvals_knw_W].min() < 0:
                    raise ValueError("All finite elements of `known_W` must be >= 0.")
        if known_H is not None:
            mask_fvals_knw_H = np.isfinite(known_H)
            tot_fvals_knw_H = np.sum(mask_fvals_knw_H)
            if tot_fvals_knw_H == 0:
                warnings.warn(
                    "`known_H` contains no finite values, setting it to None."
                )
                known_H = None
            elif tot_fvals_knw_H == known_H.size:
                raise ValueError(
                    "All values of `H` are known, can't perform NMF. Try `NNLS solver` "
                    "to compute W."
                )
            else:
                # finite elems are non-neg
                if known_H[mask_fvals_knw_H].min() < 0:
                    raise ValueError("All finite elements of `known_H` must be >= 0.")


class FroALS(_BaseNMF):
    """
    Frobenius norm-based Nonnegative Matrix Factorization (NMF) using Alternating Least
    Squares (ALS) method.

    Factorizes a nonnegative matrix :math:`X` into two nonnegative matrices :math:`W`
    and :math:`H` by minimizing the squared Frobenius norm between :math:`X` and product
    :math:`WH` using FPGM algorithm. The following objective function is minimized
    subject to nonnegativity and other optional constraints on :math:`W` and :math:`H`:

    .. math::
        f_{\\textrm{obj}} = ||X - WH||_F^2

    where :math:`||\\cdot||_{F}` is the Frobenius norm.

    Parameters
    ----------
    rank : int
        The number of components for the factorization.
    constraint_kind : integer-like {0, 1, 2, 3, 4}, default=0
        The following constraints are applied based on the integer value specified:

        * If ``0``: Only :math:`W \\geq 0`, :math:`H \\geq 0`.
        * If ``1``: Closure constraint :math:`H^T e ≤ e`.
        * If ``2``: Closure constraint :math:`H e = e`.
        * If ``3``: Constraint :math:`W^T e = e`.
        * If ``4``: Closure constraint :math:`H^T e = e`.

        Note, for 1, 2, 3, and 4 values of ``constraint_kind`` nonnegativity constraints
        are also applied along with the additional constraint specified above.
    unimodal : dict or None, default=None
        Specifies unimodality constraints for :math:`W` and :math:`H` matrices. If
        ``None``, no unimodality constraints are applied.

        If ``dict``, Format: {'W': bool | list of bool, 'H': bool | list of bool}:

            * Must contain at least one key ('W' or 'H')
            * No other keys besides 'W' and 'H' are allowed
            * For 'W': Controls unimodality of columns in W
            * For 'H': Controls unimodality of rows in H

        Each value can be:

            * A boolean: applies to all components
            * A list of booleans: selectively applies to specific components

        Examples:

            * ``{'H': True}``: All :math:`H` components are unimodal
            * ``{'W': True, 'H': True}``: All :math:`W` and :math:`H` components are
              unimodal
            * ``{'H': [True, False, True]}``: Only components 0 and 2 of :math:`H` have
              unimodal behavior
            * ``{'W': [False, True, False]}``: Only component 1 of :math:`W` has
              unimodal behavior
    iter_max : int, default=500
        Maximum number of iterations. It must be greater :math:`\\geq 10`.
    tol : float, default=1e-4
        Tolerance for convergence. Must be in the interval :math:`(0, 1)`.

            Convergence is reached when:

            .. math::

                {|e[i] - e[i-10]| \\over e[i]} \\leq \\textrm{tol}

            where:

            * iteration :math:`i \\geq 10`
            * :math:`e[i]` is the squared relative loss after iteration :math:`i`, which
              is defined as

                .. math::

                    e[i] = {||X - W^{i}H^{i}||_{F}^2 \\over ||X||_{F}^2}

                where :math:`W^{i}` and :math:`H^{i}` is the value of :math:`W` and
                :math:`H`, respectively, after iteration :math:`i`.

    References
    ----------
    .. [1] Gillis, Nicolas. Nonnegative matrix factorization. Society for Industrial and
           Applied Mathematics, 2020.
    .. [2] Van Benthem, Mark H., and Michael R. Keenan. "Fast algorithm for the solution
           of large‐scale non‐negativity‐constrained least squares problems." Journal of
           Chemometrics: A Journal of the Chemometrics Society 18.10 (2004): 441-450.

    Examples
    --------
    >>> from mcrnmf.models import FroALS, SNPA
    >>> from mcrnmf.datasets import load_rxn_spectra
    >>>
    >>> # load the example dataset from mcrnmf
    >>> X, wv, time  = load_rxn_spectra()
    >>>
    >>> # generate initial guess using SNPA
    >>> snpa = SNPA(rank=4)
    >>> snpa.fit(X)
    >>> Wi = snpa.W  # Initial estimate for W
    >>> Hi = snpa.H  # Initial estimate for H
    >>>
    >>> # create an instance of FroALS and fit the model
    >>> model = FroALS(rank=4, constraint_kind=1, iter_max=2000, tol=1e-4)
    >>> model.fit(X, Wi, Hi)
    >>> # access decomposed factors
    >>> W, H = model.W, model.H
    >>> # check convergence status
    >>> converged = model.is_converged
    >>> # access rel. reconstruction error after each iterations
    >>> rel_recon_err = model.rel_reconstruction_error_ls
    """

    def __init__(
        self,
        rank: int,
        constraint_kind: int = 0,
        unimodal: dict | None = None,
        iter_max: int = 500,
        tol: int = 1e-4,
    ):
        super().__init__(
            rank=rank,
            constraint_kind=constraint_kind,
            unimodal=unimodal,
            iter_max=iter_max,
            tol=tol,
        )

    @property
    def rel_reconstruction_error_ls(self):
        """
        List of relative reconstruction errors from each iteration during model fitting.

        The relative reconstruction error measures how well the current factors
        approximate the original data. It is the ratio:

        .. math::
            \\dfrac{||X - W^{i}H^{i}||_F}{||X||_F}\\ \\textrm{,}

        where:

        * :math:`||\\cdot||_F` denotes the Frobenius norm
        * :math:`X` is the original data matrix
        * :math:`W^{i}` and :math:`H^{i}` are values of :math:`W` and :math:`H` after
          iteration :math:`i`

        Returns
        -------
        list of float
            Relative reconstruction error from each iteration.
        """
        if self._rel_loss_ls is None:
            raise AttributeError(
                "`rel_reconstruction_error_ls` is not available. Model has not been "
                "fitted yet. Call 'fit' first."
            )
        return self._rel_loss_ls

    def fit(
        self,
        X: NDArray[np.float64],
        Wi: NDArray[np.float64],
        Hi: NDArray[np.float64],
        known_W: NDArray[np.float64] | None = None,
        known_H: NDArray[np.float64] | None = None,
        preprocess_scale_WH: bool = False,
    ):
        """
        Fit the FroALS model to the provided data.

        Parameters
        ----------
        X : ndarray of shape (n_features, n_samples)
            Data array to be factorized.
        Wi : ndarray of shape (n_features, rank)
            Initial guess for the factor :math:`W`.
        Hi : ndarray of shape (rank, n_samples)
            Initial guess for the factor :math:`H`.
        known_W : ndarray of shape (n_features, rank), default=None
            Array containing known values of :math:`W`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`W` which do not
              correspond ``np.nan`` entries in ``known_W``.
        known_H : ndarray of shape (rank, n_samples), default=None
            Array containing known values of :math:`H`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`H` which do not
              correspond ``np.nan`` entries in ``known_H``.
        preprocess_scale_WH : bool, default=False
            If ``True``,  ``Wi`` and ``Hi`` are scaled before optimization.
        """
        self._validate_fit_args(X, Wi, Hi, known_W, known_H)
        self._W = Wi.copy()
        self._H = Hi.copy()
        self._rel_loss_ls = []
        self._is_converged = False

        # Check if known values are passed, if so compute a mask.
        if known_W is not None:
            mask_known_W = np.isfinite(known_W)
        else:
            mask_known_W = None
        if known_H is not None:
            mask_known_H = np.isfinite(known_H)
        else:
            mask_known_H = None
        # preallocate arrays for H updates
        WtW = np.zeros((self.rank_, self.rank_))
        WtX = np.zeros((self.rank_, X.shape[1]))

        # preallocate arrays for W updates
        XHt = X @ self._H.T
        HHt = self._H @ self._H.T
        if preprocess_scale_WH is True:
            self._W[:] = self._W * (
                np.sum(XHt * self._W) / (np.sum((self._W.T @ self._W) * HHt))
            )
            norm_Wcol = np.linalg.norm(self._W, axis=0) + _EPSILON
            norm_Hrow = np.linalg.norm(self._H, axis=1) + _EPSILON
            scaling_vec = np.sqrt(norm_Wcol / norm_Hrow)
            self._W[:] = self._W / scaling_vec
            self._H[:] = self._H * scaling_vec[:, np.newaxis]
            HHt[:] = HHt * scaling_vec[np.newaxis, :]
            HHt[:] = scaling_vec[:, np.newaxis] * HHt
            XHt[:] = XHt * scaling_vec[np.newaxis, :]

        norm_X = np.linalg.norm(X, ord="fro")
        rel_loss = np.sqrt(self._fro_err_sq_after_W_update(norm_X, HHt, XHt)) / norm_X
        self._rel_loss_ls.append(rel_loss)

        iter = 0
        while iter < self.iter_max_:
            self._update_H(X, WtW, WtX, known_H, mask_known_H)
            self._update_W(X, HHt, XHt, known_W, mask_known_W)
            rel_loss = (
                np.sqrt(self._fro_err_sq_after_W_update(norm_X, HHt, XHt)) / norm_X
            )
            self._rel_loss_ls.append(rel_loss)
            self._is_converged = self._check_convergence(iter)
            if self._is_converged:
                break
            iter += 1

    def _update_H(
        self,
        X: NDArray[np.float64],
        WtW: NDArray[np.float64],
        WtX: NDArray[np.float64],
        known_H: NDArray[np.float64] | None,
        mask_known_H: NDArray[np.bool_] | None,
    ):
        """
        Update the H matrix based on the current W matrix using ALS.

        Solves the least squares problem H = argmin ||X - WH||_F^2 subject to
        constraints. Uses efficient matrix operations and handles potential
        ill-conditioning.

        Parameters
        ----------
        X : ndarray
            The input matrix to be factorized.
        WtW : ndarray
            Precomputed W^T * W matrix for efficiency.
        WtX : ndarray
            Precomputed W^T * X matrix for efficiency.
        known_H : ndarray or None
            Matrix of known values in H (NaN elements are treated as unknown).
        mask_known_H : ndarray or None
            Boolean mask indicating which values in H are known.

        Returns
        -------
        None
            The H matrix is updated in-place.
        """
        WtW[:] = self._W.T @ self._W
        WtX[:] = self._W.T @ X
        if np.linalg.cond(WtW) > 1e6:
            delta = np.trace(WtW) / self.rank_
            WtW[:] += 1e-6 * delta * np.eye(self.rank_)
            self._H[:] = np.linalg.solve(WtW, WtX)
        else:
            self._H[:] = np.linalg.solve(WtW, WtX)
        # apply known values
        if mask_known_H is not None:
            self._H[mask_known_H] = known_H[mask_known_H]
        self._apply_unimodal_H_rows()
        self._project_H()

    def _update_W(
        self,
        X: NDArray[np.float64],
        HHt: NDArray[np.float64],
        XHt: NDArray[np.float64],
        known_W: NDArray[np.float64] | None,
        mask_known_W: NDArray[np.bool_] | None,
    ):
        """
        Update the W matrix based on the current H matrix using ALS.

        Solves the least squares problem W = argmin ||X - WH||_F^2 subject to
        constraints. Uses efficient matrix operations and handles potential
        ill-conditioning.

        Parameters
        ----------
        X : ndarray
            The input matrix to be factorized.
        HHt : ndarray
            Precomputed H * H^T matrix for efficiency.
        XHt : ndarray
            Precomputed X * H^T matrix for efficiency.
        known_W : ndarray or None
            Matrix of known values in W (NaN elements are treated as unknown).
        mask_known_W : ndarray or None
            Boolean mask indicating which values in W are known.

        Returns
        -------
        None
            The W matrix is updated in-place.
        """
        HHt[:] = self._H @ self._H.T
        XHt[:] = X @ self._H.T
        if np.linalg.cond(HHt) > 1e6:
            delta = np.trace(HHt) / self.rank_
            HHt[:] += 1e-6 * delta * np.eye(self.rank_)
            self._W[:] = np.linalg.solve(HHt, XHt.T).T
        else:
            self._W[:] = np.linalg.solve(HHt, XHt.T).T
        # apply known values
        if mask_known_W is not None:
            self._W[mask_known_W] = known_W[mask_known_W]
        self._apply_unimodal_W_cols()
        self._project_W()


class FroFPGM(_BaseNMF):
    """
    Frobenius norm-based Nonnegative Matrix Factorization (NMF) using Fast Projected
    Gradient Method (FPGM).

    Factorizes a nonnegative matrix :math:`X` into two nonnegative matrices :math:`W`
    and :math:`H` by minimizing the squared Frobenius norm between :math:`X` and product
    :math:`WH` using FPGM algorithm. The following objective function is minimized
    subject to nonnegativity and other optional constraints on :math:`W` and :math:`H`:

    .. math::
        f_{\\textrm{obj}} = ||X - WH||_F^2

    where :math:`||\\cdot||_{F}` is the Frobenius norm.

    Parameters
    ----------
    rank : int
        The number of components for the factorization.
    constraint_kind : integer-like {0, 1, 2, 3, 4}, default=0
        The following constraints are applied based on the integer value specified:

        * If ``0``: Only :math:`W \\geq 0`, :math:`H \\geq 0`.
        * If ``1``: Closure constraint :math:`H^T e ≤ e`.
        * If ``2``: Closure constraint :math:`H e = e`.
        * If ``3``: Constraint :math:`W^T e = e`.
        * If ``4``: Closure constraint :math:`H^T e = e`.

        Note, for 1, 2, 3, and 4 values of ``constraint_kind`` nonnegativity constraints
        are also applied along with the additional constraint specified above.
    unimodal : dict or None, default=None
        Specifies unimodality constraints for :math:`W` and :math:`H` matrices. If
        ``None``, no unimodality constraints are applied.

        If ``dict``, Format: {'W': bool | list of bool, 'H': bool | list of bool}:

            * Must contain at least one key ('W' or 'H')
            * No other keys besides 'W' and 'H' are allowed
            * For 'W': Controls unimodality of columns in W
            * For 'H': Controls unimodality of rows in H

        Each value can be:

            * A boolean: applies to all components
            * A list of booleans: selectively applies to specific components

        Examples:

            * ``{'H': True}``: All :math:`H` components are unimodal
            * ``{'W': True, 'H': True}``: All :math:`W` and :math:`H` components are
              unimodal
            * ``{'H': [True, False, True]}``: Only components 0 and 2 of :math:`H` have
              unimodal behavior
            * ``{'W': [False, True, False]}``: Only component 1 of :math:`W` has
              unimodal behavior
    iter_max : int, default=500
        Maximum number of iterations. It must be greater :math:`\\geq 10`.
    tol : float, default=1e-4
        Tolerance for convergence. Must be in the interval :math:`(0, 1)`.

        Convergence is reached when:

            .. math::

                {|e[i] - e[i-10]| \\over e[i]} \\leq \\textrm{tol}

            where:

            * iteration :math:`i \\geq 10`
            * :math:`e[i]` is the squared relative loss after iteration :math:`i`, which
              is defined as

                .. math::

                    e[i] = {||X - W^{i}H^{i}||_{F}^2 \\over ||X||_{F}^2}

                where :math:`W^{i}` and :math:`H^{i}` is the value of :math:`W` and
                :math:`H`, respectively, after iteration :math:`i`.
    inner_iter_max : int, default=20
        Maximum number of inner iterations performed during each single update of either
        :math:`W` or :math:`H` while the other is held fixed.
    inner_iter_tol : float, default=0.1
        Tolerance for the convergence of the inner loop. The inner loop convergence is
        reached when:

        .. math::

            {||A^{k} - A^{k-1}||_{F} \\over ||A^{1} - A^{0}||_{F}} \\leq
            \\textrm{inner_iter_tol}

        where:

        * :math:`A` is either :math:`H` or :math:`W`.
        * :math:`A^{k}` is value of :math:`A` after inner-iteration :math:`k`.
        * :math:`A^{0}` is the value of :math:`A` before the first inner-iteration.

    References
    ----------
    .. [1] Gillis, N. (2020). Nonnegative matrix factorization. Society for Industrial
           and Applied Mathematics.
    .. [2] Lin, Chih-Jen. "Projected gradient methods for nonnegative matrix
           factorization." Neural computation 19.10 (2007): 2756-2779.

    Examples
    --------
    >>> from mcrnmf.models import FroFPGM, SNPA
    >>> from mcrnmf.datasets import load_rxn_spectra
    >>>
    >>> # load the example dataset from mcrnmf
    >>> X, wv, time  = load_rxn_spectra()
    >>>
    >>> # generate initial guess using SNPA
    >>> snpa = SNPA(rank=4)
    >>> snpa.fit(X)
    >>> Wi = snpa.W  # Initial estimate for W
    >>> Hi = snpa.H  # Initial estimate for H
    >>>
    >>> # create an instance of FroALS and fit the model
    >>> model = FroFPGM(rank=4, constraint_kind=1, iter_max=2000, tol=1e-4)
    >>> model.fit(X, Wi, Hi)
    >>> # access decomposed factors
    >>> W, H = model.W, model.H
    >>> # check convergence status
    >>> converged = model.is_converged
    >>> # access rel. reconstruction error after each iterations
    >>> rel_recon_err = model.rel_reconstruction_error_ls
    """

    def __init__(
        self,
        rank: int,
        constraint_kind: int = 0,
        unimodal: dict | None = None,
        iter_max: int = 500,
        tol: int = 0.0001,
        inner_iter_max: int = 20,
        inner_iter_tol: float = 0.1,
    ):
        super().__init__(
            rank=rank,
            constraint_kind=constraint_kind,
            unimodal=unimodal,
            iter_max=iter_max,
            tol=tol,
        )

        if not isinstance(inner_iter_max, (int, np.integer)):
            raise TypeError(
                f"`inner_iter_max` must be of type {int}, current type is "
                f"{type(inner_iter_max)}"
            )
        if not isinstance(inner_iter_tol, float):
            raise TypeError(
                f"`tol` must be of type {float}, current type is {type(inner_iter_tol)}"
            )

        if inner_iter_max <= 0:
            raise ValueError("`inner_iter_max` must be > 0")
        if not (0 < inner_iter_tol < 1):
            raise ValueError("`inner_iter_tol` must be in the open interval (0, 1)")

        self.inner_iter_max_ = inner_iter_max
        self.inner_iter_tol_ = inner_iter_tol

    @property
    def rel_reconstruction_error_ls(self):
        """
        List of relative reconstruction errors from each iteration during model fitting.

        It is defined as:

        .. math::
            \\dfrac{||X - W^{i}H^{i}||_F}{||X||_F}\\ \\textrm{,}

        where:

        * :math:`||\\cdot||_F` denotes the Frobenius norm
        * :math:`X` is the original data matrix
        * :math:`W^{i}` and :math:`H^{i}` are values of :math:`W` and :math:`H` after
          iteration :math:`i`

        Returns
        -------
        list of float
            Relative reconstruction error from each iteration.
        """
        if self._rel_loss_ls is None:
            raise AttributeError(
                "`rel_reconstruction_error_ls` is not available. Model has not been "
                "fitted yet. Call `fit` first."
            )
        return self._rel_loss_ls

    def fit(
        self,
        X: NDArray[np.float64],
        Wi: NDArray[np.float64],
        Hi: NDArray[np.float64],
        known_W: NDArray[np.float64] | None = None,
        known_H: NDArray[np.float64] | None = None,
        preprocess_scale_WH: bool = False,
    ):
        """
        Fit the FroFPGM model to the provided data.

        Parameters
        ----------
        X : ndarray of shape (n_features, n_samples)
            Data array to be factorized.
        Wi : ndarray of shape (n_features, rank)
            Initial guess for the factor :math:`W`.
        Hi : ndarray of shape (rank, n_samples)
            Initial guess for the factor :math:`H`.
        known_W : ndarray of shape (n_features, rank), default=None
            Array containing known values of :math:`W`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`W` which do not
              correspond ``np.nan`` entries in ``known_W``.
        known_H : ndarray of shape (rank, n_samples), default=None
            Array containing known values of :math:`H`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`H` which do not
              correspond ``np.nan`` entries in ``known_H``.
        preprocess_scale_WH : bool, default=False
            If ``True``,  ``Wi`` and ``Hi`` are scaled before optimization.
        """
        self._validate_fit_args(X, Wi, Hi, known_W, known_H)
        self._W = Wi.copy()
        self._H = Hi.copy()
        self._rel_loss_ls = []
        self._is_converged = False

        # Check if known values are passed, if so compute a mask.
        if known_W is not None:
            mask_known_W = np.isfinite(known_W)
        else:
            mask_known_W = None
        if known_H is not None:
            mask_known_H = np.isfinite(known_H)
        else:
            mask_known_H = None

        # preallocate arrays for H updates
        WtW = np.zeros((self.rank_, self.rank_))
        WtX = np.zeros((self.rank_, X.shape[1]))
        Y_H = np.zeros_like(self._H)
        H_diff = np.zeros_like(self._H)
        H_prev = np.zeros_like(self._H)

        # preallocate arrays for W updates
        XHt = X @ self._H.T
        HHt = self._H @ self._H.T
        Y_W = np.zeros_like(self._W)
        W_diff = np.zeros_like(self._W)
        W_prev = np.zeros_like(self._W)

        if preprocess_scale_WH is True:
            self._W[:] = self._W * (
                np.sum(XHt * self._W) / (np.sum((self._W.T @ self._W) * HHt))
            )
            norm_Wcol = np.linalg.norm(self._W, axis=0) + _EPSILON
            norm_Hrow = np.linalg.norm(self._H, axis=1) + _EPSILON
            scaling_vec = np.sqrt(norm_Wcol / norm_Hrow)
            self._W[:] = self._W / scaling_vec
            self._H[:] = self._H * scaling_vec[:, np.newaxis]
            HHt[:] = HHt * scaling_vec[np.newaxis, :]
            HHt[:] = scaling_vec[:, np.newaxis] * HHt
            XHt[:] = XHt * scaling_vec[np.newaxis, :]

        norm_X = np.linalg.norm(X, ord="fro")
        rel_loss = np.sqrt(self._fro_err_sq_after_W_update(norm_X, HHt, XHt)) / norm_X
        self._rel_loss_ls.append(rel_loss)

        iter = 0
        while iter < self.iter_max_:
            self._update_H(X, WtW, WtX, Y_H, H_diff, H_prev, known_H, mask_known_H)
            self._update_W(X, HHt, XHt, Y_W, W_diff, W_prev, known_W, mask_known_W)
            rel_loss = (
                np.sqrt(self._fro_err_sq_after_W_update(norm_X, HHt, XHt)) / norm_X
            )
            self._rel_loss_ls.append(rel_loss)
            self._is_converged = self._check_convergence(iter)
            if self._is_converged:
                break
            iter += 1

    def _update_H(
        self,
        X: NDArray[np.float64],
        WtW: NDArray[np.float64],
        WtX: NDArray[np.float64],
        Y_H: NDArray[np.float64],
        H_diff: NDArray[np.float64],
        H_prev: NDArray[np.float64],
        known_H: None | NDArray[np.float64],
        mask_known_H: None | NDArray[np.float64],
    ) -> None:
        """
        Update the H matrix based on the current W matrix.

        This method implements the FPGM update for the H matrix.

        Parameters
        ----------
        X : ndarray
            The input matrix to be factorized.
        WtW : ndarray
            Precomputed W^T * W matrix.
        WtX : ndarray
            Precomputed W^T * X matrix.
        Y_H : ndarray
            Auxiliary variable for momentum updates.
        H_diff : ndarray
            Difference between consecutive H updates.
        H_prev : ndarray
            Previous H matrix for momentum calculation.
        known_H : ndarray or None
            Matrix with known values of H.
        mask_known_H : ndarray or None
            Mask indicating which values in H are known.

        Returns
        -------
        None
            The H matrix is updated in-place.
        """
        self._project_H()
        WtW[:] = self._W.T @ self._W
        WtX[:] = self._W.T @ X

        # hyperparameters for optimization
        l = np.linalg.norm(WtW, 2)  # lipschitz constant
        theta_prev = 0.05  # first value of theta, heuristic

        Y_H = self._H.copy()  # look ahead variable
        inner_iter = 0
        while inner_iter < self.inner_iter_max_:
            H_prev[:] = self._H.copy()
            self._H[:] = Y_H - ((WtW @ Y_H) - WtX) / l
            # apply known values
            if mask_known_H is not None:
                self._H[mask_known_H] = known_H[mask_known_H]
            self._apply_unimodal_H_rows()
            self._project_H()
            H_diff[:] = self._H - H_prev
            norm_H_diff = np.linalg.norm(H_diff, "fro")
            if inner_iter == 0:
                norm_H_diff_0 = norm_H_diff
            # convergence check
            if norm_H_diff <= (self.inner_iter_tol_ * norm_H_diff_0):
                break
            # do the momentum update only if convergence criteria is not satisfied
            theta_prev_sq = theta_prev**2
            # theta and beta computation based on scheme 2.2.19 in Introductory lectures
            # on convex optimization by Nesterov
            theta = (np.sqrt(theta_prev_sq**2 + 4 * theta_prev_sq) - theta_prev_sq) / 2
            beta = theta_prev * (1 - theta_prev) / (theta_prev_sq + theta)
            # momentum step
            H_diff[:] = H_diff * beta
            Y_H[:] = self._H + H_diff
            theta_prev = theta
            inner_iter += 1

    def _update_W(
        self,
        X: NDArray[np.float64],
        HHt: NDArray[np.float64],
        XHt: NDArray[np.float64],
        Y_W: NDArray[np.float64],
        W_diff: NDArray[np.float64],
        W_prev: NDArray[np.float64],
        known_W: None | NDArray[np.float64],
        mask_known_W: None | NDArray[np.float64],
    ) -> None:
        """
        Update the W matrix based on the current H matrix.

        This method implements the FPGM update for the W matrix.

        Parameters
        ----------
        X : ndarray
            The input matrix to be factorized.
        HHt : ndarray
            Precomputed H * H^T matrix.
        XHt : ndarray
            Precomputed X * H^T matrix.
        Y_W : ndarray
            Auxiliary variable for momentum updates.
        W_diff : ndarray
            Difference between consecutive W updates.
        W_prev : ndarray
            Previous W matrix for momentum calculation.
        known_W : ndarray or None
            Matrix with known values of W.
        mask_known_W : ndarray or None
            Mask indicating which values in W are known.

        Returns
        -------
        None
            The W matrix is updated in-place.
        """
        self._project_W()
        HHt[:] = self._H @ self._H.T
        XHt[:] = X @ self._H.T

        # hyperparameters for optimization
        l = np.linalg.norm(HHt, 2)  # lipschitz constant
        theta_prev = 0.05  # first value of theta, heuristic

        Y_W = self._W.copy()  # look ahead variable
        inner_iter = 0
        while inner_iter < self.inner_iter_max_:
            W_prev[:] = self._W.copy()
            self._W[:] = Y_W - ((Y_W @ HHt) - XHt) / l
            # apply known values
            if mask_known_W is not None:
                self._W[mask_known_W] = known_W[mask_known_W]
            self._apply_unimodal_W_cols()
            self._project_W()
            W_diff[:] = self._W - W_prev
            norm_W_diff = np.linalg.norm(W_diff, "fro")
            if inner_iter == 0:
                norm_W_diff_0 = norm_W_diff
            # convergence check
            if norm_W_diff <= (self.inner_iter_tol_ * norm_W_diff_0):
                break
            # do the momentum update only if convergence criteria is not satisfied
            theta_prev_sq = theta_prev**2
            # theta and beta computation based on scheme 2.2.19 in Introductory lectures
            # on convex optimization by Nesterov
            theta = (np.sqrt(theta_prev_sq**2 + 4 * theta_prev_sq) - theta_prev_sq) / 2
            beta = theta_prev * (1 - theta_prev) / (theta_prev_sq + theta)
            # momentum step
            W_diff[:] = W_diff * beta
            Y_W[:] = self._W + W_diff
            theta_prev = theta
            inner_iter += 1


class MinVol(FroFPGM):
    """
    Minimum-Volume Nonnegative Matrix Factorization (NMF) implementation using Fast
    Projected Gradient Method.

    This class implements the FPGM to perform Minimum-Volume NMF, minimizing the
    following objective function subject to non-negativity and other optional
    constraints.

    .. math::
        f_{\\textrm{obj}} = ||X - WH||_F^2 + \\lambda \\times \\log(\\det(W^TW + \\delta
        \\times I))

    where:

    * :math:`||\\cdot||_{F}` is the Frobenius norm.
    * :math:`X` the nonnegative matrix to be factorized
    * :math:`W` and :math:`H` are the nonnegative factors
    * :math:`\\lambda` and :math:`\\delta` are model parameters. See definition in the
      Parameters section
    * :math:`I` is an identity matrix

    Parameters
    ----------
    rank : int
        The number of components for the factorization.
    constraint_kind : integer-like {0, 1, 2, 3, 4}, default=0
        The following constraints are applied based on the integer value specified:

        * If ``0``: Only :math:`W \\geq 0`, :math:`H \\geq 0`.
        * If ``1``: Closure constraint :math:`H^T e ≤ e`.
        * If ``2``: Closure constraint :math:`H e = e`.
        * If ``3``: Constraint :math:`W^T e = e`.
        * If ``4``: Closure constraint :math:`H^T e = e`.

        Note, for 1, 2, 3, and 4 values of ``constraint_kind`` nonnegativity constraints
        are also applied along with the additional constraint specified above.
    unimodal : dict or None, default=None
        Specifies unimodality constraints for :math:`W` and :math:`H` matrices. If
        ``None``, no unimodality constraints are applied.

        If ``dict``, Format: {'W': bool | list of bool, 'H': bool | list of bool}:

            * Must contain at least one key ('W' or 'H')
            * No other keys besides 'W' and 'H' are allowed
            * For 'W': Controls unimodality of columns in W
            * For 'H': Controls unimodality of rows in H

        Each value can be:

            * A boolean: applies to all components
            * A list of booleans: selectively applies to specific components

        Examples:

            * ``{'H': True}``: All :math:`H` components are unimodal
            * ``{'W': True, 'H': True}``: All :math:`W` and :math:`H` components are
              unimodal
            * ``{'H': [True, False, True]}``: Only components 0 and 2 of :math:`H` have
              unimodal behavior
            * ``{'W': [False, True, False]}``: Only component 1 of :math:`W` has
              unimodal behavior
    iter_max : int, default=500
        Maximum number of iterations. It must be greater :math:`\\geq 10`.
    tol : float, default=1e-4
        Tolerance for convergence. Must be in the interval :math:`(0, 1)`.

            Convergence is reached when:

            .. math::

                {|e[i] - e[i-10]| \\over e[i]} \\leq \\textrm{tol}

            where:

            * iteration :math:`i \\geq 10`
            * :math:`e[i]` is the squared relative loss after iteration :math:`i`, which
              is defined as

                .. math::

                    e[i] = {||X - W^{i}H^{i}||_{F}^2 \\over ||X||_{F}^2}

                where :math:`W^{i}` and :math:`H^{i}` is the value of :math:`W` and
                :math:`H`, respectively, after iteration :math:`i`.
    inner_iter_max : int, default=20
        Maximum number of inner iterations performed during each single update of either
        :math:`W` or :math:`H` while the other is held fixed.
    inner_iter_tol : float, default=0.1
        Tolerance for the convergence of the inner loop while updating :math:`H`.

            The inner loop convergence is reached when:

            .. math::

                {||H^{k} - H^{k-1}||_{F} \\over ||H^{1} - H^{0}||_{F}} \\leq
                \\textrm{inner_iter_tol}

            where:

            * :math:`H^{k}` is value of :math:`H` after inner-iteration :math:`k`.
            * :math:`H^{0}` is the value of :math:`H` before the first inner-iteration.
    delta : float, default=0.1
        Value of parameter :math:`\\delta`. It ensures numerical stability when
        :math:`W` is rank-deficient.
    lambdaa : float, default=1e-3
        The intial weight, :math:`\\lambda`, of the volume-regularization term.

    References
    ----------
    .. [1] Gillis, N. (2020). Nonnegative matrix factorization. Society for Industrial
           and Applied Mathematics.
    .. [2] Leplat, Valentin, Andersen MS Ang, and Nicolas Gillis. "Minimum-volume rank-
           deficient nonnegative matrix factorizations." ICASSP 2019-2019 IEEE
           International Conference on Acoustics, Speech and Signal Processing (ICASSP).
           IEEE, 2019.

    Examples
    --------
    >>> from mcrnmf.models import MinVol, SNPA
    >>> from mcrnmf.datasets import load_rxn_spectra
    >>>
    >>> # load the example dataset from mcrnmf
    >>> X, wv, time  = load_rxn_spectra()
    >>>
    >>> # generate initial guess using SNPA
    >>> snpa = SNPA(rank=4)
    >>> snpa.fit(X)
    >>> Wi = snpa.W  # Initial estimate for W
    >>> Hi = snpa.H  # Initial estimate for H
    >>>
    >>> # create an instance of FroALS and fit the model
    >>> model = MinVol(rank=4, constraint_kind=1, iter_max=2000, tol=1e-4, lambdaa=1e-3)
    >>> model.fit(X, Wi, Hi)
    >>> # access decomposed factors
    >>> W, H = model.W, model.H
    >>> # check convergence status
    >>> converged = model.is_converged
    >>> # access rel. reconstruction error after each iterations
    >>> rel_recon_err = model.rel_reconstruction_error_ls
    """

    def __init__(
        self,
        rank: int,
        constraint_kind: int = 0,
        unimodal: dict | None = None,
        iter_max: int = 500,
        tol: float = 0.0001,
        inner_iter_max: int = 20,
        inner_iter_tol: float = 0.1,
        delta: float | int = 0.1,
        lambdaa: float = 0.001,
    ):
        super().__init__(
            rank=rank,
            constraint_kind=constraint_kind,
            unimodal=unimodal,
            iter_max=iter_max,
            tol=tol,
            inner_iter_max=inner_iter_max,
            inner_iter_tol=inner_iter_tol,
        )
        if not isinstance(delta, (int, np.integer, float)):
            raise TypeError(
                f"`delta` must be of type {float} or {int}, current type is "
                f"{type(delta)}"
            )
        if not isinstance(lambdaa, (float, int, np.integer)):
            raise TypeError(
                f"`lambdaa` must be of type {float} or {int}, current type is "
                f"{type(lambdaa)}"
            )
        if delta <= 0:
            raise ValueError("`delta` must be > 0")
        if lambdaa <= 0:
            raise ValueError("`lambdaa` must be > 0")

        self.delta_ = delta
        self.lambdaa_ = lambdaa

        # initialize the additional attribute specific to MinVol
        self._rel_reconstruction_error_ls = None

    @property
    def rel_reconstruction_error_ls(self):
        """
        List of relative reconstruction errors from each iteration during model fitting.

        The relative reconstruction error measures how well the current factors
        approximate the original data. It is the ratio:

        .. math::
            \\dfrac{||X - W^{i}H^{i}||_F}{||X||_F}\\ \\textrm{,}

        where:

        * :math:`||\\cdot||_F` denotes the Frobenius norm
        * :math:`X` is the original data matrix
        * :math:`W^{i}` and :math:`H^{i}` are values of :math:`W` and :math:`H` after
          iteration :math:`i`

        Returns
        -------
        list of float
            Relative reconstruction error from each iteration.
        """
        if self._rel_reconstruction_error_ls is None:
            raise AttributeError(
                "`rel_reconstruction_error_ls` is not available. Model has not been "
                "fitted yet. Call `fit` first."
            )
        return self._rel_reconstruction_error_ls

    def fit(
        self,
        X: NDArray[np.floating],
        Wi: NDArray[np.floating],
        Hi: NDArray[np.floating],
        known_W: None | NDArray[np.floating] = None,
        known_H: None | NDArray[np.floating] = None,
        preprocess_scale_WH: bool = True,
    ):
        """
        Fit the MinVol model to the provided data.

        Parameters
        ----------
        X : ndarray of shape (n_features, n_samples)
            Data array to be factorized.
        Wi : ndarray of shape (n_features, rank)
            Initial guess for the factor :math:`W`.
        Hi : ndarray of shape (rank, n_samples)
            Initial guess for the factor :math:`H`.
        known_W : ndarray of shape (n_features, rank), default=None
            Array containing known values of :math:`W`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`W` which do not
              correspond ``np.nan`` entries in ``known_W``.
        known_H : ndarray of shape (rank, n_samples), default=None
            Array containing known values of :math:`H`.

            * The ``np.nan`` elements of the array are treated as unknown.
            * Equality constraint is applied at those indices of :math:`H` which do not
              correspond ``np.nan`` entries in ``known_H``.
        preprocess_scale_WH : bool, default=True
            If ``True``,  ``Wi`` and ``Hi`` are scaled before optimization.
        """
        self._validate_fit_args(X, Wi, Hi, known_W, known_H)
        self._W = Wi.copy()
        self._H = Hi.copy()
        self._rel_loss_ls = []
        self._rel_reconstruction_error_ls = []
        self._is_converged = False

        # Check if known values are passed, if so compute a mask.
        if known_W is not None:
            mask_known_W = np.isfinite(known_W)
        else:
            mask_known_W = None
        if known_H is not None:
            mask_known_H = np.isfinite(known_H)
        else:
            mask_known_H = None

        # pre-allocate matrices used in update H
        Y_H = np.zeros_like(self._H)
        H_diff = np.zeros_like(self._H)
        H_prev = np.zeros_like(self._H)
        WtW = np.zeros((self.rank_, self.rank_))
        WtX = np.zeros((self.rank_, X.shape[1]))

        # preallocate matrices used in update W
        Y_W = np.zeros_like(self._W)
        W_diff = np.zeros_like(self._W)
        W_prev = np.zeros_like(self._W)
        HHt = self._H @ self._H.T
        XHt = X @ self._H.T

        # scale W and H based on the constraint
        if (self.constraint_kind_ != 0) and preprocess_scale_WH:
            self._scale_WH(
                X, H_prev, Y_W, W_diff, W_prev, HHt, XHt, known_W, mask_known_W
            )

        norm_X = np.sqrt(np.sum(X**2))
        WtW[:] = self._W.T @ self._W
        WtX[:] = self._W.T @ X
        I = np.eye(self.rank_, dtype=X.dtype)
        fro_error_sq = self._fro_err_sq_after_H_update(norm_X, WtW, WtX)
        self._rel_reconstruction_error_ls.append(np.sqrt(fro_error_sq) / norm_X)
        vol_error = self._calc_vol_error(WtW=WtW, I=I)
        lambdaa_updt = self.lambdaa_ * max(1e-6, fro_error_sq) / np.abs(vol_error)
        rel_loss = np.sqrt(fro_error_sq + lambdaa_updt * vol_error) / norm_X
        self._rel_loss_ls.append(rel_loss)

        # pre-allocate memory for frequently used matrices
        A = np.zeros((self.rank_, self.rank_))
        Y = np.zeros((self.rank_, self.rank_))
        iter = 0
        while iter < self.iter_max_:
            XHt[:] = X @ self._H.T
            HHt[:] = self._H @ self._H.T
            Y[:] = np.linalg.inv(WtW + self.delta_ * I)
            A[:] = lambdaa_updt * Y + HHt
            self._update_W_minvol(A, XHt, Y_W, W_diff, W_prev, known_W, mask_known_W)
            self._update_H(X, WtW, WtX, Y_H, H_diff, H_prev, known_H, mask_known_H)
            fro_error_sq = self._fro_err_sq_after_H_update(norm_X, WtW, WtX)
            self._rel_reconstruction_error_ls.append(np.sqrt(fro_error_sq) / norm_X)
            vol_error = self._calc_vol_error(WtW=WtW, I=I)
            rel_loss = np.sqrt(fro_error_sq + lambdaa_updt * vol_error) / norm_X
            self._rel_loss_ls.append(rel_loss)
            self._is_converged = self._check_convergence(iter)
            if self._is_converged:
                break
            iter += 1

    def _calc_vol_error(
        self, WtW: NDArray[np.float64], I: NDArray[np.float64]
    ) -> np.float64:
        """
        Calculate the volume error term in the objective function.

        Computes log(det(W^T * W + delta * I)) which quantifies the volume of the
        simplex defined by the columns of W.

        Parameters
        ----------
        WtW : ndarray
            Precomputed W^T·W matrix.
        I : ndarray
            Identity matrix of appropriate size.

        Returns
        -------
        float
            The volume error term.
        """
        return np.log(np.linalg.det(WtW + self.delta_ * I))

    def _scale_WH(
        self,
        X: NDArray[np.float64],
        H_prev: NDArray[np.float64],
        Y_W: NDArray[np.float64],
        W_diff: NDArray[np.float64],
        W_prev: NDArray[np.float64],
        HHt: NDArray[np.float64],
        XHt: NDArray[np.float64],
        known_W: None | NDArray[np.float64],
        mask_known_W: None | NDArray[np.float64],
    ) -> None:
        """
        Scale W and H matrices to maintain constraints while improving conditioning.

        Depending on the constraint type, this method applies appropriate scaling
        transformations to W and H that preserve their product.

        Parameters
        ----------
        X : ndarray
            The input matrix.
        H_prev : ndarray
            Previous H matrix for comparison.
        Y_W : ndarray
            Auxiliary variable for W updates.
        W_diff : ndarray
            Difference between consecutive W updates.
        W_prev : ndarray
            Previous W matrix.
        HHt : ndarray
            Precomputed H·H^T matrix.
        XHt : ndarray
            Precomputed X·H^T matrix.
        known_W : ndarray or None
            Matrix with known values of W.
        mask_known_W : ndarray or None
            Mask indicating which values in W are known.

        Returns
        -------
        None
            The W and H matrices are updated in-place.
        """
        if self.constraint_kind_ in [1, 4]:
            H_prev = self._H.copy()
            if self.constraint_kind_ == 1:
                simplex_proj(self._H)
            else:
                simplex_proj_using_axis(self._H, True)
            # update W if the H projection has changed H significantly
            if np.linalg.norm(self._H - H_prev, 2) > (
                1e-3 * np.linalg.norm(self._H, 2)
            ):
                self._update_W(X, HHt, XHt, Y_W, W_diff, W_prev, known_W, mask_known_W)
        elif self.constraint_kind_ == 2:
            H_row_sum = np.sum(self._H, axis=1)
            self._H[:] = self._H / H_row_sum[:, np.newaxis]
            self._W[:] = self._W * H_row_sum
        elif self.constraint_kind_ == 3:
            W_col_sum = np.sum(self._W, axis=0)
            self._H[:] = self._H * W_col_sum[:, np.newaxis]
            self._W[:] = self._W / W_col_sum

    def _update_W_minvol(
        self,
        A: NDArray[np.float64],
        XHt: NDArray[np.float64],
        Y_W: NDArray[np.float64],
        W_diff: NDArray[np.float64],
        W_prev: NDArray[np.float64],
        known_W: NDArray[np.float64] | None,
        mask_known_W: NDArray[np.float64] | None,
    ) -> None:
        """
        Update W with minimum volume regularization.

        Performs gradient descent on W with the objective function that includes
        both data fidelity and volume regularization terms. The update optimizes:
        ||X - WH||_F^2 + lambdaa * log(det(W^T*W + delta*I))

        Parameters
        ----------
        A : ndarray
            Precomputed regularized Hessian matrix.
        XHt : ndarray
            Precomputed X*H^T matrix.
        Y_W : ndarray
            Look-ahead matrix for momentum updates.
        W_diff : ndarray
            Matrix to store differences between iterations.
        W_prev : ndarray
            Matrix to store previous W values.
        known_W : ndarray or None
            Matrix containing known values for W.
        mask_known_W : ndarray or None
            Boolean mask indicating which W values are known.

        Returns
        -------
        None
            The W matrix is updated in-place.
        """
        self._project_W()
        # hyperparamters for optimization
        s = np.linalg.svd(A, compute_uv=False)
        l = s[0]  # lipschitz constant
        kappa_inv = s[-1] / s[0]  # inverse of the condition number
        beta = (1 - np.sqrt(kappa_inv)) / (1 + np.sqrt(kappa_inv))

        Y_W = self._W.copy()  # look ahead variable
        inner_iter = 0
        error, error_prev = 1, 0
        while inner_iter < self.inner_iter_max_:
            W_prev[:] = self._W.copy()
            self._W[:] = Y_W - ((Y_W @ A) - XHt) / l
            # apply known values
            if mask_known_W is not None:
                self._W[mask_known_W] = known_W[mask_known_W]
            self._apply_unimodal_W_cols()
            self._project_W()
            W_diff[:] = self._W - W_prev
            norm_W_diff = np.linalg.norm(W_diff, "fro")
            if inner_iter == 0:
                norm_W_diff_0 = norm_W_diff
            # convergence check
            if norm_W_diff <= (1e-6 * norm_W_diff_0):
                break
            error = np.sum((self._W.T @ self._W) * A) - 2 * np.sum(self._W * XHt)
            if (inner_iter == 0) or (error <= error_prev):
                W_diff[:] = W_diff * beta
                Y_W[:] = self._W + W_diff
            else:
                # If current error is greater than previous error, replace look ahead
                # variable with current value of W
                Y_W[:] = self._W.copy()
            error_prev = error
            inner_iter += 1


class SNPA:
    """
    Successive Nonnegative Projection Algorithm (SNPA) for NMF initialization.

    The algorithm sequentially selects columns from :math:`X` with maximum residual norm
    after projecting onto the convex cone generated by previously selected columns,
    providing a robust initialization for :math:`W` which can be used by other NMF
    models. The corresponding :math:`H` matrix is then computed by minimizing
    :math:`||X - WH||_F^2` subject to nonnegativity and simplex constraints on
    :math:`H`.

    Parameters
    ----------
    rank : int
        The number of components to extract.
    iter_max : int, default=500
        Maximum number of iterations for updating :math:`H` after extraction of
        :math:`W`.
    tol : float, default=1e-6
        Tolerance for early stopping. The algorithm stops extracting basis compnents
        when:

        .. math::
            \\dfrac{||X - WH||_F}{||X||_{F}} \\leq \\textrm{tol}

        where :math:`||\\cdot||_{F}` is the Frobenius norm.

    Examples
    --------
    >>> from mcrnmf.models import SNPA
    >>> from mcrnmf.datasets import load_rxn_spectra
    >>>
    >>> # load the example dataset from mcrnmf
    >>> X, wv, time  = load_rxn_spectra()
    >>>
    >>> # create an instance of SNPA and fit to the data
    >>> snpa = SNPA(rank=4)
    >>> snpa.fit(X)
    >>> Wi = snpa.W  # Estimate for W
    >>> Hi = snpa.H  # Estimate for H
    >>> selected_indices = model.col_indices_ls
    """

    def __init__(
        self,
        rank: int,
        iter_max: int = 500,
        tol: float = 1e-6,
    ):
        # validate types of input arguments
        if not isinstance(rank, (int, np.integer)):
            raise TypeError(
                f"`rank` must be of type {int}, currrent type is {type(rank)}"
            )
        if not isinstance(iter_max, (np.integer, int)):
            raise TypeError(
                f"`iter_max` must be of type {int}, current type is {type(iter_max)}"
            )
        if not isinstance(tol, float):
            raise TypeError(
                f"`tol` must be of type {float}, current type is {type(tol)}"
            )
        # validate values of input arguments
        if rank < 1:
            raise ValueError("`rank` must be > 0")
        if iter_max < 1:
            raise ValueError("`iter_max` must be > 0")
        if not (0 < tol < 1):
            raise ValueError("`tol` must be in the open interval (0, 1)")

        self.rank_ = rank
        self.iter_max_ = iter_max
        self.tol_ = tol

        # Initialize the attributes
        self._W = None
        self._H = None
        self._col_indices_ls = None

    @property
    def W(self):
        """
        The basis matrix containing selected columns from :math:`X`.

        Returns
        -------
        ndarray of shape (n_features, rank)
            The basis matrix containing selected columns from :math:`X`.
        """
        if self._W is None:
            raise AttributeError(
                "'W' is not available. Model has not been fitted yet. Call 'fit' first."
            )
        return self._W

    @property
    def H(self):
        """
        The coefficient matrix derived from :math:`W`.

        Returns
        -------
        ndarray of shape (rank, n_samples)
            The coefficient matrix derived from :math:`W`.
        """
        if self._H is None:
            raise AttributeError(
                "'H' is not available. Model has not been fitted yet. Call 'fit' first."
            )
        return self._H

    @property
    def col_indices_ls(self):
        """
        The indices of columns selected from :math:`X` for :math:`W`.

        Returns
        -------
        list of int
            The indices of columns selected from :math:`X` for :math:`W` after fitting
            the model.
        """
        if self._col_indices_ls is None:
            raise AttributeError(
                "'col_indices_ls' is not available. Model has not been fitted yet. "
                "Call 'fit' first."
            )
        return self._col_indices_ls

    def fit(
        self,
        X: NDArray[np.float64],
    ):
        """
        Fit the SNPA model to the provided data.

        Parameters
        ----------
        X : ndarray of shape (n_features, n_samples)
            Data array to be factorized.

        Warns
        -----
        UserWarning
            If the desired rank is not reached due to early convergence. If this happens
            try specifying a lower value for ``tol``.
        """
        # validate X
        if not isinstance(X, np.ndarray):
            raise TypeError("`X` must be a numpy array")
        if X.dtype != np.float64:
            raise TypeError(
                f"The dtype of `X` elements should be {np.float64}, current dtype is "
                f"{X.dtype}"
            )
        if self.rank_ > X.shape[1]:
            raise ValueError(
                f"`rank` cannot be greater than columns in `X` {X.shape[1]}."
            )
        if X.min() < 0:
            raise ValueError("All elements of `X` must be >= 0")

        X_col_norm = np.sum(X**2, axis=0)
        X_col_norm_max = np.max(X_col_norm)

        # initialize Residuals column norm
        R_col_norm = X_col_norm.copy()

        self._col_indices_ls = []  # a list to store the column indices
        XtW = np.empty(0)
        self._W = np.empty(0)
        self._H = np.empty(0)
        idx_count = 0
        while idx_count < self.rank_:
            R_col_norm_max = np.max(R_col_norm)
            # check if there are multiples columns with max value
            idx = np.where((R_col_norm_max - R_col_norm) / R_col_norm_max <= 1e-6)[0]
            if idx.size > 1:
                # in case of ties select the one of with largest col norm
                self._col_indices_ls.append(idx[np.argmax(X_col_norm[idx])])
            else:
                self._col_indices_ls.append(idx[0])

            self._W = X[:, self._col_indices_ls]
            Wi = self._W[:, idx_count].reshape(-1, 1)
            if XtW.size == 0:
                XtW = X.T @ self._W
            else:
                XtW = np.hstack((XtW, X.T @ Wi))

            if idx_count == 0:
                WiTWi = self._W.T @ self._W
            else:
                WtWi = self._W[:, :idx_count].T @ Wi
                WiTWi = np.vstack(
                    (np.hstack((WiTWi, WtWi)), np.hstack((WtWi.T, Wi.T @ Wi)))
                )
            if idx_count == 0:
                self._update_H(X)
            else:
                self._H[:, self._col_indices_ls[-1]] = 0
                h = np.zeros((1, X.shape[1]))
                h[0, self._col_indices_ls[-1]] = 1
                self._H = np.vstack([self._H, h])
                self._update_H(X)

            if idx_count == 0:
                R_col_norm[:] = (
                    X_col_norm - 2 * (XtW.T * self._H) + (self._H * (WiTWi @ self._H))
                )
            else:
                R_col_norm[:] = (
                    X_col_norm
                    - 2 * np.sum(XtW.T * self._H, axis=0)
                    + np.sum(self._H * (WiTWi @ self._H), axis=0)
                )
            np.maximum(R_col_norm, 0, out=R_col_norm)

            if np.sqrt(np.max(R_col_norm) / X_col_norm_max) <= self.tol_:
                if len(self._col_indices_ls) != self.rank_:
                    warnings.warn(
                        f"Extracted only {len(self._col_indices_ls)} of {self.rank_} "
                        "indices"
                    )
                break
            idx_count += 1

    def _update_H(self, X: NDArray[np.float64]):
        """
        Update the coefficient matrix H given data X and basis W.

        This method solves the optimization problem to find H that minimizes
        ||X - WH||_F subject to simplex constraints on H. It uses an accelerated
        projected gradient descent approach with momentum updates for faster
        convergence.

        Parameters
        ----------
        X : ndarray
            The input data matrix of shape (n_features, n_samples).

        Notes
        -----
        This is an internal method used by the `fit` method and should not be
        called directly by users.
        """
        # initialization of H while doing rank-1 factorization
        if self._H.size == 0:
            if np.linalg.cond(self._W) > 1e6:
                warnings.warn(
                    "`W` has condition number > 1e6. Initial guess of H generated might"
                    " be poor."
                )
            self._H = np.linalg.lstsq(self._W, X, rcond=None)[0]
            self._H[:] = np.maximum(self._H, 0.0)
            alpha = np.sum(self._H * (self._W.T @ X)) / np.sum(
                (self._W.T @ self._W) * (self._H @ self._H.T)
            )
            self._H[:] = self._H * alpha

            # replaces rows which sum to zero with small random values
            zero_sum_rows = np.where(np.sum(self._H, axis=1) == 0)[0]
            if len(zero_sum_rows) > 0:
                self._H[zero_sum_rows, :] = (
                    0.001
                    * np.max(self._H)
                    * np.random.rand(len(zero_sum_rows), self._H.shape[1])
                )
        simplex_proj(self._H)

        # precompute WtW, and WtX since it doesn't change in the loop
        WtW = self._W.T @ self._W
        WtX = self._W.T @ X

        # hyperparameters for optimization
        l = np.linalg.norm(WtW, 2)  # lipschitz constant
        alpha_prev = 0.05  # first value of alpha, heuristic
        delta = 1e-6

        # pre-allocate memory for matrices computed or used repeatedly in the loop
        Y = self._H.copy()  # look ahead variable
        H_diff = np.zeros(self._H.shape)
        H_prev = np.zeros(self._H.shape)

        iter = 0
        while iter < self.iter_max_:
            H_prev[:] = self._H.copy()
            self._H[:] = Y - ((WtW @ Y) - WtX) / l
            simplex_proj(self._H)
            H_diff[:] = self._H - H_prev
            norm_H_diff = np.linalg.norm(H_diff, "fro")
            if iter == 0:
                norm_H_diff_0 = norm_H_diff
            # convergence check
            if norm_H_diff <= (delta * norm_H_diff_0):
                break
            # do the momentum update only if convergence criteria is not satisfied
            alpha_prev_sq = alpha_prev**2
            alpha = (np.sqrt(alpha_prev_sq**2 + 4 * alpha_prev_sq) - alpha_prev_sq) / 2
            beta = alpha_prev * (1 - alpha_prev) / (alpha_prev_sq + alpha)
            # momentum step
            H_diff[:] = H_diff * beta
            Y[:] = self._H + H_diff
            alpha_prev = alpha
            iter += 1
